'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
__doc__ = """
The ``pm4py.ml`` module contains the machine learning features offered in ``pm4py``.
"""

from typing import Union, Tuple, Any, List, Collection, Optional
import pandas as pd
import numpy as np
from pm4py.objects.ocel.obj import OCEL
from pm4py.objects.log.obj import EventLog
from pm4py.utils import __event_log_deprecation_warning
import random
from copy import copy
from pm4py.util.pandas_utils import (
    check_is_pandas_dataframe,
    check_pandas_dataframe_columns,
)
from pm4py.utils import get_properties, constants, pandas_utils, is_polars_lazyframe, check_is_pandas_dataframe


def split_train_test(
    log: Union[EventLog, pd.DataFrame],
    train_percentage: float = 0.8,
    case_id_key: str = "case:concept:name",
) -> Union[Tuple[EventLog, EventLog], Tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Splits an event log into a training log and a test log for machine learning purposes.

    This function separates the provided log into two parts based on the specified training percentage.
    It ensures that entire cases are included in either the training set or the test set.

    :param log: The event log or Pandas DataFrame to be split.
    :param train_percentage: Fraction of cases to be included in the training log (between 0.0 and 1.0).
    :param case_id_key: Attribute to be used as the case identifier.
    :return: A tuple containing the training and test event logs or DataFrames.
    :rtype: ``Union[Tuple[EventLog, EventLog], Tuple[pd.DataFrame, pd.DataFrame]]``

    .. code-block:: python3

        import pm4py

        train_df, test_df = pm4py.split_train_test(dataframe, train_percentage=0.75)
    """
    __event_log_deprecation_warning(log)

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log)
        cases = pandas_utils.format_unique(log[case_id_key].unique())
        train_cases = set()
        test_cases = set()
        for c in cases:
            r = random.random()
            if r <= train_percentage:
                train_cases.add(c)
            else:
                test_cases.add(c)
        train_df = log[log[case_id_key].isin(train_cases)]
        test_df = log[log[case_id_key].isin(test_cases)]
        return train_df, test_df
    else:
        from pm4py.objects.log.util import split_train_test

        return split_train_test.split(log, train_percentage=train_percentage)


def get_prefixes_from_log(
    log: Union[EventLog, pd.DataFrame],
    length: int,
    case_id_key: str = "case:concept:name",
) -> Union[EventLog, pd.DataFrame]:
    """
    Retrieves prefixes of traces in a log up to a specified length.

    The returned log contains prefixes of each trace:
    - If a trace has a length less than or equal to the specified length, it is included as-is.
    - If a trace exceeds the specified length, it is truncated to that length.

    :param log: The event log or Pandas DataFrame from which to extract prefixes.
    :param length: The maximum length of prefixes to extract.
    :param case_id_key: Attribute to be used as the case identifier.
    :return: A log containing the prefixes of the original log.
    :rtype: ``Union[EventLog, pd.DataFrame]``

    .. code-block:: python3

        import pm4py

        trimmed_df = pm4py.get_prefixes_from_log(dataframe, length=5, case_id_key='case:concept:name')
    """
    __event_log_deprecation_warning(log)

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, case_id_key=case_id_key)
        from pm4py.util import pandas_utils

        log = pandas_utils.insert_ev_in_tr_index(log, case_id=case_id_key)
        return log[log[constants.DEFAULT_INDEX_IN_TRACE_KEY] <= (length - 1)]
    else:
        from pm4py.objects.log.util import get_prefixes

        return get_prefixes.get_prefixes_from_log(log, length)


def extract_outcome_enriched_dataframe(
    log: Union[EventLog, pd.DataFrame],
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
    start_timestamp_key: str = "time:timestamp",
) -> pd.DataFrame:
    """
    Enriches a dataframe with additional outcome-related columns computed from the entire case.

    This function adds columns that model the outcome of each case by computing metrics such as
    arrival rates and service waiting times.

    :param log: The event log or Pandas DataFrame to be enriched.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as the case identifier.
    :param start_timestamp_key: Attribute to be used as the start timestamp.
    :return: An enriched Pandas DataFrame with additional outcome-related columns.
    :rtype: ``pd.DataFrame``

    .. code-block:: python3

        import pm4py

        enriched_df = pm4py.extract_outcome_enriched_dataframe(
            log,
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name',
            start_timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    properties = get_properties(
        log,
        activity_key=activity_key,
        case_id_key=case_id_key,
        timestamp_key=timestamp_key,
    )

    if not check_is_pandas_dataframe(log):
        from pm4py.objects.conversion.log import converter as log_converter
        log = log_converter.apply(
            log,
            variant=log_converter.Variants.TO_DATA_FRAME,
            parameters=properties,
        )

    log = copy(log)
    fea_df = extract_features_dataframe(
        log,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
        include_case_id=True,
    )
    log = pandas_utils.insert_case_arrival_finish_rate(
        log,
        timestamp_column=timestamp_key,
        case_id_column=case_id_key,
        start_timestamp_column=start_timestamp_key,
    )
    log = pandas_utils.insert_case_service_waiting_time(
        log,
        timestamp_column=timestamp_key,
        case_id_column=case_id_key,
        start_timestamp_column=start_timestamp_key,
    )

    if is_polars_lazyframe(log):
        return log.join(fea_df, on=case_id_key)
    else:
        return log.merge(fea_df, on=case_id_key)


def extract_features_dataframe(
    log: Union[EventLog, pd.DataFrame],
    str_tr_attr: Optional[List[str]] = None,
    num_tr_attr: Optional[List[str]] = None,
    str_ev_attr: Optional[List[str]] = None,
    num_ev_attr: Optional[List[str]] = None,
    str_evsucc_attr: Optional[List[str]] = None,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: Optional[str] = None,
    resource_key: str = "org:resource",
    include_case_id: bool = False, count_occurrences: bool = False,
    **kwargs
) -> pd.DataFrame:
    """
    Extracts a dataframe containing features for each case in the provided log object.

    This function processes the log to generate a set of features that can be used for machine learning tasks.
    Features can include both case-level and event-level attributes, with options for one-hot encoding.

    :param log: The event log or Pandas DataFrame from which to extract features.
    :param str_tr_attr: (Optional) List of string attributes at the case level to extract as features.
    :param num_tr_attr: (Optional) List of numeric attributes at the case level to extract as features.
    :param str_ev_attr: (Optional) List of string attributes at the event level to extract as features (one-hot encoded).
    :param num_ev_attr: (Optional) List of numeric attributes at the event level to extract as features
                       (uses the last value per attribute in a case).
    :param str_evsucc_attr: (Optional) List of string successor attributes at the event level to extract as features.
    :param activity_key: Attribute to be used as the activity identifier.
    :param timestamp_key: Attribute to be used for timestamps.
    :param case_id_key: (Optional) Attribute to be used as the case identifier. If not provided, the default is used.
    :param resource_key: Attribute to be used as the resource identifier.
    :param include_case_id: Whether to include the case identifier column in the features table.
    :param **kwargs: Additional keyword arguments to pass to the feature extraction algorithm.
    :return: A Pandas DataFrame containing the extracted features for each case.
    :rtype: ``pd.DataFrame``

    .. code-block:: python3

        import pm4py

        features_df = pm4py.extract_features_dataframe(
            dataframe,
            activity_key='concept:name',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = kwargs if kwargs else {}

    properties = get_properties(
        log,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
    )
    for prop in properties:
        parameters[prop] = properties[prop]

    parameters["str_tr_attr"] = str_tr_attr or []
    parameters["num_tr_attr"] = num_tr_attr or []
    parameters["str_ev_attr"] = str_ev_attr or []
    parameters["num_ev_attr"] = num_ev_attr or []
    parameters["str_evsucc_attr"] = str_evsucc_attr or []
    parameters["add_case_identifier_column"] = include_case_id
    parameters["count_occurrences"] = count_occurrences

    from pm4py.algo.transformation.log_to_features import (
        algorithm as log_to_features,
    )

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log,
            activity_key=activity_key,
            case_id_key=case_id_key,
            timestamp_key=timestamp_key,
        )

    data, feature_names = log_to_features.apply(log, parameters=parameters)
    if is_polars_lazyframe(data):
        return data

    return pandas_utils.instantiate_dataframe(data, columns=feature_names)


def extract_ocel_features(
    ocel: OCEL,
    obj_type: str,
    enable_object_lifecycle_paths: bool = True,
    enable_object_work_in_progress: bool = False,
    object_str_attributes: Optional[Collection[str]] = None,
    object_num_attributes: Optional[Collection[str]] = None,
    include_obj_id: bool = False,
    debug: bool = False,
) -> pd.DataFrame:
    """
    Extracts a set of features from an object-centric event log (OCEL) for objects of a specified type.

    This function computes various features based on the lifecycle paths and work-in-progress metrics
    of objects within the OCEL. It also supports encoding of string and numeric object attributes.

    The approach is based on:
    Berti, A., Herforth, J., Qafari, M.S. et al. Graph-based feature extraction on object-centric event logs.
    Int J Data Sci Anal (2023). https://doi.org/10.1007/s41060-023-00428-2

    :param ocel: The object-centric event log from which to extract features.
    :param obj_type: The object type to consider for feature extraction.
    :param enable_object_lifecycle_paths: Whether to enable the "lifecycle paths" feature.
    :param enable_object_work_in_progress: Whether to enable the "work in progress" feature,
                                           which has a high computational cost.
    :param object_str_attributes: (Optional) Collection of string attributes at the object level to one-hot encode.
    :param object_num_attributes: (Optional) Collection of numeric attributes at the object level to encode.
    :param include_obj_id: Whether to include the object identifier as a column in the features DataFrame.
    :param debug: Whether to enable debugging mode to track the feature extraction process.
    :return: A Pandas DataFrame containing the extracted features for the specified object type.
    :rtype: ``pd.DataFrame``

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('log.jsonocel')
        fea_df = pm4py.extract_ocel_features(ocel, "item")
    """
    if object_str_attributes is None:
        object_str_attributes = []

    if object_num_attributes is None:
        object_num_attributes = []

    parameters = {
        "filter_per_type": obj_type,
        "enable_object_lifecycle_paths": enable_object_lifecycle_paths,
        "enable_object_work_in_progress": enable_object_work_in_progress,
        "enable_object_str_attributes": bool(object_str_attributes),
        "enable_object_num_attributes": bool(object_num_attributes),
        "str_obj_attr": object_str_attributes,
        "num_obj_attr": object_num_attributes,
        "debug": debug,
    }

    from pm4py.algo.transformation.ocel.features.objects import (
        algorithm as ocel_feature_extraction,
    )

    data, feature_names = ocel_feature_extraction.apply(
        ocel, parameters=parameters
    )

    dataframe = pandas_utils.instantiate_dataframe(data, columns=feature_names)
    dataframe = dataframe.dropna(how="any", axis=1)
    dataframe = dataframe.select_dtypes(include=np.number)

    if include_obj_id:
        objects_with_type = ocel.objects[
            [ocel.object_id_column, ocel.object_type_column]
        ].to_dict("records")
        objects_with_type = [
            x[ocel.object_id_column]
            for x in objects_with_type
            if x[ocel.object_type_column] == obj_type
        ]
        dataframe[ocel.object_id_column] = objects_with_type

    return dataframe


def extract_temporal_features_dataframe(
    log: Union[EventLog, pd.DataFrame],
    grouper_freq: str = "W",
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: Optional[str] = None,
    start_timestamp_key: str = "time:timestamp",
    resource_key: str = "org:resource",
) -> pd.DataFrame:
    """
    Extracts temporal features from a log object and returns them as a dataframe.

    This function computes temporal metrics based on the specified grouping frequency, which can be
    daily (D), weekly (W), monthly (M), or yearly (Y). These features are useful for analyzing
    system dynamics and simulation in the context of process mining.

    The approach is based on:
    Pourbafrani, Mahsa, Sebastiaan J. van Zelst, and Wil MP van der Aalst.
    "Supporting automatic system dynamics model generation for simulation in the context of process mining."
    International Conference on Business Information Systems. Springer, Cham, 2020.

    :param log: The event log or Pandas DataFrame from which to extract temporal features.
    :param grouper_freq: The frequency to use for grouping (e.g., 'D' for daily, 'W' for weekly,
                         'M' for monthly, 'Y' for yearly).
    :param activity_key: Attribute to be used as the activity identifier.
    :param timestamp_key: Attribute to be used for timestamps.
    :param case_id_key: (Optional) Attribute to be used as the case identifier. If not provided, the default is used.
    :param start_timestamp_key: Attribute to be used as the start timestamp.
    :param resource_key: Attribute to be used as the resource identifier.
    :return: A Pandas DataFrame containing the extracted temporal features.
    :rtype: ``pd.DataFrame``

    .. code-block:: python3

        import pm4py

        temporal_features_df = pm4py.extract_temporal_features_dataframe(
            dataframe,
            activity_key='concept:name',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(
        log,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
    )

    parameters["grouper_freq"] = grouper_freq
    parameters["pm4py:param:activity_key"] = activity_key
    parameters["pm4py:param:timestamp_key"] = timestamp_key
    if case_id_key is not None:
        parameters["pm4py:param:case_id_key"] = case_id_key
    parameters["pm4py:param:start_timestamp_key"] = (
        start_timestamp_key
    )
    parameters["pm4py:param:resource_key"] = resource_key

    if is_polars_lazyframe(log):
        from pm4py.algo.transformation.log_to_features.variants import temporal_lazy as temporal
    else:
        from pm4py.algo.transformation.log_to_features.variants import temporal

    return temporal.apply(log, parameters=parameters)


def extract_target_vector(
    log: Union[EventLog, pd.DataFrame],
    variant: str,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
) -> Tuple[Any, List[str]]:
    """
    Extracts the target vector from a log object for a specific machine learning use case.

    Supported variants include:
    - 'next_activity': Predicts the next activity in a case.
    - 'next_time': Predicts the timestamp of the next activity.
    - 'remaining_time': Predicts the remaining time for the case.

    :param log: The event log or Pandas DataFrame from which to extract the target vector.
    :param variant: The variant of the algorithm to use. Must be one of:
                    'next_activity', 'next_time', 'remaining_time'.
    :param activity_key: Attribute to be used as the activity identifier.
    :param timestamp_key: Attribute to be used for timestamps.
    :param case_id_key: Attribute to be used as the case identifier.
    :return: A tuple containing the target vector and a list of class labels (if applicable).
    :rtype: ``Tuple[Any, List[str]]``

    :raises Exception: If an unsupported variant is provided.

    .. code-block:: python3

        import pm4py

        vector_next_act, class_next_act = pm4py.extract_target_vector(
            log,
            'next_activity',
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
        vector_next_time, class_next_time = pm4py.extract_target_vector(
            log,
            'next_time',
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
        vector_rem_time, class_rem_time = pm4py.extract_target_vector(
            log,
            'remaining_time',
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(
        log,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
    )

    from pm4py.algo.transformation.log_to_target import (
        algorithm as log_to_target,
    )

    var_map = {
        "next_activity": log_to_target.Variants.NEXT_ACTIVITY,
        "next_time": log_to_target.Variants.NEXT_TIME,
        "remaining_time": log_to_target.Variants.REMAINING_TIME,
    }

    if variant not in var_map:
        raise Exception(
            "Please provide the variant as one of the following: 'next_activity', 'next_time', 'remaining_time'."
        )

    target, classes = log_to_target.apply(
        log, variant=var_map[variant], parameters=parameters
    )
    return target, classes



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum

from pm4py.algo.discovery.footprints import algorithm as fp_discovery
from pm4py.objects.log.obj import EventLog, Trace, Event
from pm4py.objects.process_tree.utils import bottomup as bottomup_discovery
from pm4py.objects.process_tree.obj import Operator
from pm4py.util import constants, xes_constants
from pm4py.util import exec_utils

from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog
from pm4py.objects.process_tree.obj import ProcessTree

TRACES = "traces"
SKIPPABLE = "skippable"


class Parameters(Enum):
    MIN_TRACE_LENGTH = "min_trace_length"
    MAX_TRACE_LENGTH = "max_trace_length"
    MAX_LOOP_OCC = "max_loop_occ"
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    MAX_LIMIT_NUM_TRACES = "max_limit_num_traces"
    RETURN_SET_STRINGS = "return_set_strings"


def get_playout_leaf(
    node,
    playout_dictio,
    min_trace_length,
    max_trace_length,
    max_loop_occ,
    min_rem_dict,
    max_rem_dict,
    max_limit_num_traces,
):
    """
    Performs the playout of a leaf (activity or invisible), returning the traces  allowed by the tree
    """
    mr = min_rem_dict[node]
    mar = max_rem_dict[node]
    playout_dictio[node] = {TRACES: set()}
    if node.label is None:
        playout_dictio[node][TRACES].add(tuple([]))
    else:
        if mar + 1 >= min_trace_length and max_trace_length - mr >= 1:
            playout_dictio[node][TRACES].add((node.label,))


def get_playout_xor(
    node,
    playout_dictio,
    min_trace_length,
    max_trace_length,
    max_loop_occ,
    min_rem_dict,
    max_rem_dict,
    max_limit_num_traces,
):
    """
    Performs the playout of a XOR node, returning the traces allowed by the tree
    """
    mr = min_rem_dict[node]
    traces = set()
    for n in node.children:
        traces = traces.union(playout_dictio[n][TRACES])
        if len(traces) > max_limit_num_traces:
            break
    playout_dictio[node] = {TRACES: traces}


def get_min_remaining_length(traces):
    """
    Minimum remaining length (for sequential, parallel cut detection)

    Parameters
    --------------
    traces
        Traces
    """
    min_len_traces = []
    min_rem_length = []
    for x in traces:
        if len(x) == 0:
            min_len_traces.append(0)
        else:
            min_len_traces.append(len(x[0]))
        min_rem_length.append(0)
    min_rem_length[-1] = 0
    min_rem_length[-2] = min_len_traces[-1]
    j = len(traces) - 3
    while j >= 0:
        min_rem_length[j] = min_rem_length[j + 1] + min_len_traces[j + 1]
        j = j - 1
    return min_len_traces, min_rem_length


def get_max_remaining_length(traces):
    """
    Maximum remaining length (for sequential, parallel cut detection)

    Parameters
    --------------
    traces
        Traces
    """
    max_len_traces = []
    max_rem_length = []
    for x in traces:
        if len(x) == 0:
            max_len_traces.append(0)
        else:
            max_len_traces.append(len(x[-1]))
        max_rem_length.append(0)
    max_rem_length[-1] = 0
    max_rem_length[-2] = max_len_traces[-1]
    j = len(traces) - 3
    while j >= 0:
        max_rem_length[j] = max_rem_length[j + 1] + max_len_traces[j + 1]
        j = j - 1
    return max_len_traces, max_rem_length


def flatten(x):
    """
    Flattens a list of tuples
    """
    ret = []
    for y in x:
        for z in y:
            ret.append(z)
    return ret


def get_sequential_compositions_children(
    traces, min_trace_length, max_trace_length, mr, mar, max_limit_num_traces
):
    """
    Returns alls the possible sequential combinations between
    the children of a tree
    """
    diff = max_trace_length - mr
    diff2 = min_trace_length - mar
    min_len_traces, min_rem_length = get_min_remaining_length(traces)
    max_len_traces, max_rem_length = get_max_remaining_length(traces)
    curr = list(traces[0])
    i = 1
    while i < len(traces):
        mrl = min_rem_length[i]
        marl = max_rem_length[i]
        to_visit = []
        j = 0
        while j < len(curr):
            x = curr[j]
            if not type(x) is list:
                x = [x]
            z = 0
            while z < len(traces[i]):
                y = traces[i][z]
                xy = list(x)
                xy.append(y)
                val = sum(len(k) for k in xy)
                if val + mrl <= diff and val + marl >= diff2:
                    to_visit.append(xy)
                z = z + 1
            j = j + 1
        curr = to_visit
        i = i + 1
    return curr


def get_playout_parallel(
    node,
    playout_dictio,
    min_trace_length,
    max_trace_length,
    max_loop_occ,
    min_rem_dict,
    max_rem_dict,
    max_limit_num_traces,
):
    """
    Performs the playout of an AND node, returning the traces allowed by the tree
    """
    mr = min_rem_dict[node]
    mar = max_rem_dict[node]
    traces = list(
        sorted(playout_dictio[x][TRACES], key=lambda x: len(x))
        for x in node.children
    )
    sequential_compositions = get_sequential_compositions_children(
        traces,
        min_trace_length,
        max_trace_length,
        mr,
        mar,
        max_limit_num_traces,
    )
    final_traces = list()
    for x in sequential_compositions:
        if len(final_traces) >= max_limit_num_traces:
            break
        to_visit = [[[]] + [len(y) for y in x]]
        while len(to_visit) > 0:
            curr = to_visit.pop(0)
            possible_choices = [
                i - 1 for i in range(1, len(curr)) if curr[i] > 0
            ]
            for j in possible_choices:
                new = list(curr)
                new[0] = list(new[0])
                new[0].append(x[j][len(x[j]) - curr[j + 1]])
                new[j + 1] = new[j + 1] - 1
                to_visit.append(new)
            if not possible_choices:
                final_traces.append(tuple(curr[0]))
                if len(final_traces) >= max_limit_num_traces:
                    break
    playout_dictio[node] = {TRACES: set(final_traces)}


def get_playout_sequence(
    node,
    playout_dictio,
    min_trace_length,
    max_trace_length,
    max_loop_occ,
    min_rem_dict,
    max_rem_dict,
    max_limit_num_traces,
):
    """
    Performs the playout of a sequence node, returning the traces allowed by the tree
    """
    mr = min_rem_dict[node]
    mar = max_rem_dict[node]
    final_traces = set()
    traces = list(
        sorted(playout_dictio[x][TRACES], key=lambda x: len(x))
        for x in node.children
    )
    sequential_compositions = get_sequential_compositions_children(
        traces,
        min_trace_length,
        max_trace_length,
        mr,
        mar,
        max_limit_num_traces,
    )
    for x in sequential_compositions:
        final_traces.add(tuple(flatten(x)))
    playout_dictio[node] = {TRACES: final_traces}


def get_playout_loop(
    node,
    playout_dictio,
    min_trace_length,
    max_trace_length,
    max_loop_occ,
    min_rem_dict,
    max_rem_dict,
    max_limit_num_traces,
):
    """
    Performs the playout of a loop node, returning the traces allowed by the tree
    """
    mr = min_rem_dict[node]
    mar = max_rem_dict[node]
    final_traces = set()
    do_traces = sorted(
        list(playout_dictio[node.children[0]][TRACES]), key=lambda x: len(x)
    )
    redo_traces = sorted(
        list(playout_dictio[node.children[1]][TRACES]), key=lambda x: len(x)
    )
    min_do_trace = min(len(x) for x in do_traces) if do_traces else 0
    to_visit = list((x, 0, 0) for x in do_traces)
    closed = set()
    diff1 = max_trace_length - mr
    diff2 = max_trace_length - min_do_trace - mr
    diff3 = min_trace_length - mar
    while to_visit:
        curr = to_visit.pop(0)
        curr_trace = curr[0]
        position = curr[1]
        num_loops = curr[2]
        if position == 0:
            if curr_trace in closed:
                continue
            closed.add(curr_trace)

            if diff3 <= len(curr_trace) <= diff1:
                final_traces.add(curr_trace)
                if len(final_traces) > max_limit_num_traces:
                    break

            for y in redo_traces:
                new = curr_trace + y
                if len(new) <= diff2 and num_loops + 1 <= max_loop_occ:
                    to_visit.append((new, 1, num_loops + 1))
                else:
                    break

        elif position == 1:
            for y in do_traces:
                new = curr_trace + y
                if len(new) <= diff1:
                    to_visit.append((new, 0, num_loops))
                else:
                    break
    playout_dictio[node] = {TRACES: final_traces}


def get_playout(
    node,
    playout_dictio,
    min_trace_length,
    max_trace_length,
    max_loop_occ,
    min_rem_dict,
    max_rem_dict,
    max_limit_num_traces,
):
    """
    Performs a playout of an ode of the process tree, given the type
    """
    if len(node.children) == 0:
        get_playout_leaf(
            node,
            playout_dictio,
            min_trace_length,
            max_trace_length,
            max_loop_occ,
            min_rem_dict,
            max_rem_dict,
            max_limit_num_traces,
        )
    elif node.operator == Operator.XOR:
        get_playout_xor(
            node,
            playout_dictio,
            min_trace_length,
            max_trace_length,
            max_loop_occ,
            min_rem_dict,
            max_rem_dict,
            max_limit_num_traces,
        )
    elif node.operator == Operator.PARALLEL:
        get_playout_parallel(
            node,
            playout_dictio,
            min_trace_length,
            max_trace_length,
            max_loop_occ,
            min_rem_dict,
            max_rem_dict,
            max_limit_num_traces,
        )
    elif node.operator == Operator.SEQUENCE:
        get_playout_sequence(
            node,
            playout_dictio,
            min_trace_length,
            max_trace_length,
            max_loop_occ,
            min_rem_dict,
            max_rem_dict,
            max_limit_num_traces,
        )
    elif node.operator == Operator.LOOP:
        get_playout_loop(
            node,
            playout_dictio,
            min_trace_length,
            max_trace_length,
            max_loop_occ,
            min_rem_dict,
            max_rem_dict,
            max_limit_num_traces,
        )


def apply(
    tree: ProcessTree,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Performs an extensive playout of the process tree

    Parameters
    -------------
    tree
        Process tree
    parameters
        Possible parameters, including:
        - Parameters.MIN_TRACE_LENGTH => minimum length of a trace (default: 1)
        - Parameters.MAX_TRACE_LENGTH => maximum length of a trace (default: min_allowed_trace_length)
        - Parameters.MAX_LOOP_OCC => maximum number of occurrences for a loop (default: MAX_TRACE_LENGTH)
        - Parameters.ACTIVITY_KEY => activity key
        - Parameters.MAX_LIMIT_NUM_TRACES => maximum number to the limit of traces; the playout shall stop when the number is reached (default: 100000)
    Returns
    -------------
    log
        Event log
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    # to save memory in the returned log, allocate each activity once. to know the list of activities of the
    # process tree, use the footprints module
    fp_tree = fp_discovery.apply(tree, parameters=parameters)
    activities = fp_tree["activities"]
    activities = {act: Event({activity_key: act}) for act in activities}

    min_allowed_trace_length = bottomup_discovery.get_min_trace_length(
        tree, parameters=parameters
    )
    min_trace_length = exec_utils.get_param_value(
        Parameters.MIN_TRACE_LENGTH, parameters, 1
    )
    max_trace_length = exec_utils.get_param_value(
        Parameters.MAX_TRACE_LENGTH, parameters, min_allowed_trace_length
    )
    max_loop_occ = exec_utils.get_param_value(
        Parameters.MAX_LOOP_OCC, parameters, int(max_trace_length / 2)
    )
    max_limit_num_traces = exec_utils.get_param_value(
        Parameters.MAX_LIMIT_NUM_TRACES, parameters, 100000
    )
    return_set_strings = exec_utils.get_param_value(
        Parameters.RETURN_SET_STRINGS, parameters, False
    )

    bottomup = bottomup_discovery.get_bottomup_nodes(
        tree, parameters=parameters
    )
    min_rem_dict = bottomup_discovery.get_min_rem_dict(
        tree, parameters=parameters
    )
    max_rem_dict = bottomup_discovery.get_max_rem_dict(
        tree, parameters=parameters
    )

    playout_dictio = {}
    for i in range(len(bottomup)):
        get_playout(
            bottomup[i],
            playout_dictio,
            min_trace_length,
            max_trace_length,
            max_loop_occ,
            min_rem_dict,
            max_rem_dict,
            max_limit_num_traces,
        )
    tree_playout_traces = playout_dictio[tree][TRACES]

    if return_set_strings:
        return tree_playout_traces

    log = EventLog()
    for tr0 in tree_playout_traces:
        trace = Trace()
        for act in tr0:
            trace.append(activities[act])
        log.append(trace)

    return log



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils, constants, xes_constants
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd
from typing import Union, Dict, Optional, Any, Tuple, List
from pm4py.objects.conversion.log import converter as log_converter


class Parameters(Enum):
    ACTIVITIES = "activities"
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    ENABLE_PADDING = "enable_padding"
    PAD_SIZE = "pad_size"


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    parameters: Optional[Dict[Any, Any]] = None,
) -> Tuple[List[List[List[int]]], List[str]]:
    """
    Returns a list of matrixes (one for every case).
    Every matrix contains as many rows as many events are contained in the case (an automatic padding option is also available),
    and as many columns as many distinct activities are in the log.

    The corresponding activity to the given event is assigned to the value 1;
    the remaining activities are assigned to the value 0.

    Parameters
    --------------
    log
        Event log / Event stream / Pandas dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITIES => list of activities to consider
        - Parameters.ACTIVITY_KEY => attribute that should be used as activity
        - Parameters.ENABLE_PADDING => enables the padding (the length of cases is normalized)
        - Parameters.PAD_SIZE => the size of the padding

    Returns
    -------------
    target
        The aforementioned list of matrixes.
    activities
        The considered list of activities
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )
    max_case_length = max([len(x) for x in log])

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    activities = exec_utils.get_param_value(
        Parameters.ACTIVITIES,
        parameters,
        sorted(list(set(y[activity_key] for x in log for y in x))),
    )
    enable_padding = exec_utils.get_param_value(
        Parameters.ENABLE_PADDING, parameters, False
    )
    pad_size = exec_utils.get_param_value(
        Parameters.PAD_SIZE, parameters, max_case_length
    )

    target = []
    for trace in log:
        target.append([])
        for i in range(len(trace)):
            target[-1].append([0.0] * len(activities))
            if i < len(trace) - 1:
                act = trace[i + 1][activity_key]
                if act in activities:
                    target[-1][-1][activities.index(act)] = 1.0
        if enable_padding:
            while len(target[-1]) < pad_size:
                target[-1].append([0.0] * len(activities))

    return target, activities



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import time
import traceback
import pandas as pd
from dateutil.parser import parse
from pm4py.util.dt_parsing.variants import strpfromiso
from typing import Optional, Dict, Any
from enum import Enum
from pm4py.util import exec_utils, pandas_utils
import importlib.util


class Parameters(Enum):
    OWNER = "owner"
    REPOSITORY = "repository"
    AUTH_TOKEN = "auth_token"


def apply(parameters: Optional[Dict[Any, str]] = None) -> pd.DataFrame:
    """
    Extracts a dataframe containing the history of the issues of a Github repository.
    According to the API limit rate of public/registered users, only a part of the events
    can be returned.

    Parameters
    ---------------
    prameters
        Parameters of the algorithm, including:
        - Parameters.OWNER => owner of the repository (e.g., pm4py)
        - Parameters.REPOSITORY => name of the repository (e.g., pm4py-core)
        - Parameters.AUTH_TOKEN => authorization token

    Returns
    ---------------
    dataframe
        Pandas dataframe
    """
    import requests

    if parameters is None:
        parameters = {}

    owner = exec_utils.get_param_value(Parameters.OWNER, parameters, "pm4py")
    repo = exec_utils.get_param_value(
        Parameters.REPOSITORY, parameters, "pm4py-core"
    )
    auth_token = exec_utils.get_param_value(
        Parameters.AUTH_TOKEN, parameters, None
    )

    headers = {}
    if auth_token is not None:
        headers["Authorization"] = "Bearer " + auth_token

    continuee = True
    page = 0
    events = []

    progress = None

    while continuee:
        page += 1
        try:
            r = requests.get(
                "https://api.github.com/repos/"
                + owner
                + "/"
                + repo
                + "/issues?state=all&per_page=100&page="
                + str(page),
                headers=headers,
            )
            issues = r.json()
            if not issues:
                continuee = False
                break

            if importlib.util.find_spec("tqdm"):
                from tqdm.auto import tqdm

                progress = tqdm(
                    total=len(issues),
                    desc="extracting issues of page "
                    + str(page)
                    + ", progress :: ",
                )

            for i in issues:
                if continuee:
                    if "timeline_url" in i:
                        timeline_url = i["timeline_url"]
                        eve = {
                            "case:owner": owner,
                            "case:repo": owner + "/" + repo,
                            "case:concept:name": timeline_url,
                            "time:timestamp": strpfromiso.fix_naivety(
                                parse(i["created_at"])
                            ),
                            "concept:name": "created",
                            "org:resource": i["user"]["login"],
                            "case:author_association": i["author_association"],
                            "case:title": i["title"],
                        }
                        if "pull_request" in i:
                            eve["case:pull_request"] = i["pull_request"]["url"]
                        events.append(eve)
                        r2 = requests.get(timeline_url, headers=headers)
                        issue_events = r2.json()
                        issue_events.reverse()
                        for ev in issue_events:
                            if (
                                "created_at" in ev
                                and "event" in ev
                                and "actor" in ev
                            ):
                                eve = {
                                    "case:owner": owner,
                                    "case:repo": owner + "/" + repo,
                                    "case:concept:name": timeline_url,
                                    "time:timestamp": strpfromiso.fix_naivety(
                                        parse(ev["created_at"])
                                    ),
                                    "concept:name": ev["event"],
                                    "org:resource": ev["actor"]["login"],
                                    "case:author_association": i[
                                        "author_association"
                                    ],
                                    "case:title": i["title"],
                                }
                                if "pull_request" in i:
                                    eve["case:pull_request"] = i[
                                        "pull_request"
                                    ]["url"]
                                events.append(eve)
                        if progress is not None:
                            progress.update()
            if progress is not None:
                progress.close()
            time.sleep(1)
        except BaseException:
            continuee = False
            traceback.print_exc()
            if progress is not None:
                progress.close()
            break

    dataframe = pandas_utils.instantiate_dataframe(events)
    if len(dataframe) > 0:
        dataframe = pandas_utils.insert_index(
            dataframe, "@@index", copy_dataframe=False, reset_index=False
        )
        dataframe = dataframe.sort_values(["time:timestamp", "@@index"])
        dataframe["@@case_index"] = dataframe.groupby(
            "case:concept:name", sort=False
        ).ngroup()
        dataframe = dataframe.sort_values(
            ["@@case_index", "time:timestamp", "@@index"]
        )
    return dataframe



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union
from enum import Enum
from pm4py.objects.ocel.obj import OCEL


class Parameters(Enum):
    THETA_ACT = "theta_act"
    THETA_FLOW = "theta_flow"
    ALPHA = "alpha"
    BETA = "beta"
    GAMMA = "gamma"
    DELTA = "delta"


def apply(real: Union[OCEL, Dict[str, Any]], normative: Dict[str, Any],
          parameters: Optional[Dict[Any, Any]] = None) -> Dict[str, Any]:
    """
    Applies object-centric conformance checking between the given real object (object-centric event log or DFG)
    and a normative OC-DFG.

    Published in: https://publications.rwth-aachen.de/record/1014107

    Parameters
    -----------------
    real
        Real entity (OCEL or OC-DFG)
    normative
        Normative entity (OC-DFG)
    parameters
        Variant-specific parameters:
        - Parameters.THETA_ACT
        - Parameters.THETA_FLOW
        - Parameters.ALPHA
        - Parameters.BETA
        - Parameters.GAMMA
        - Parameters.DELTA

    Returns
    -----------------
    conf_diagn_dict
        Dictionary with conformance diagnostics
    """
    if parameters is None:
        parameters = {}

    theta_act = exec_utils.get_param_value(Parameters.THETA_ACT, parameters, 0)
    theta_flow = exec_utils.get_param_value(
        Parameters.THETA_FLOW, parameters, 0)
    alpha = exec_utils.get_param_value(Parameters.ALPHA, parameters, 1)
    beta = exec_utils.get_param_value(Parameters.BETA, parameters, 1)
    gamma = exec_utils.get_param_value(Parameters.GAMMA, parameters, 1)
    delta = exec_utils.get_param_value(Parameters.DELTA, parameters, 1)

    if isinstance(real, OCEL):
        import pm4py
        real = pm4py.discover_ocdfg(real)

    return compare_ocdfgs(
        real,
        normative,
        theta_act,
        theta_flow,
        alpha,
        beta,
        gamma,
        delta)


def compare_ocdfgs(
        ocdfg1,
        ocdfg2,
        theta_act=0,
        theta_flow=0,
        alpha=1,
        beta=1,
        gamma=1,
        delta=1):
    """
    Compare two Object-Centric Directly-Follows Graphs (OCDFGs) and perform conformance checking.

    Parameters:
    - ocdfg1: The first OCDFG to compare.
    - ocdfg2: The second OCDFG to compare.
    - theta_act: Threshold for activity measure difference.
    - theta_flow: Threshold for flow measure difference.
    - alpha, beta, gamma, delta: Weighting factors for fitness calculation.

    Returns:
    - A dictionary containing conformance checking results.
    """

    # Extract components from OCDFG1
    A1 = set(ocdfg1.get('activities', set()))
    edges1 = ocdfg1.get('edges', {})
    activities_indep1 = ocdfg1.get('activities_indep', {})

    # Extract components from OCDFG2
    A2 = set(ocdfg2.get('activities', set()))
    edges2 = ocdfg2.get('edges', {})
    activities_indep2 = ocdfg2.get('activities_indep', {})

    # Union of activities
    all_activities = A1.union(A2)

    # Activity Conformance
    A_missing = A2 - A1  # Activities in ocdfg2 but not in ocdfg1
    A_additional = A1 - A2  # Activities in ocdfg1 but not in ocdfg2

    # Flow (Edge) Conformance
    # 'event_couples' is a parent of the object types
    F1_set = set()
    event_couples1 = edges1.get('event_couples', {})
    for ot in event_couples1:
        flows1 = event_couples1[ot]
        F1_set.update(flows1.keys())

    F2_set = set()
    event_couples2 = edges2.get('event_couples', {})
    for ot in event_couples2:
        flows2 = event_couples2[ot]
        F2_set.update(flows2.keys())

    F_missing = F2_set - F1_set  # Flows in ocdfg2 but not in ocdfg1
    F_additional = F1_set - F2_set  # Flows in ocdfg1 but not in ocdfg2

    # Measure Conformance for Activities
    Delta_act = {}
    delta_act = {}
    events1 = activities_indep1.get('events', {})
    events2 = activities_indep2.get('events', {})
    for a in all_activities:
        # Measure in ocdfg1
        measure1 = len(events1.get(a, []))
        # Measure in ocdfg2
        measure2 = len(events2.get(a, []))
        diff = abs(measure2 - measure1)
        Delta_act[a] = diff
        delta_act[a] = 1 if diff > theta_act else 0

    # Measure Conformance for Flows
    Delta_flow = {}
    delta_flow = {}
    all_flows = F1_set.union(F2_set)
    for flow in all_flows:
        measure1 = 0
        # Sum over all object types in event_couples1
        for ot in event_couples1:
            flows1 = event_couples1[ot]
            measure1 += len(flows1.get(flow, []))
        measure2 = 0
        # Sum over all object types in event_couples2
        for ot in event_couples2:
            flows2 = event_couples2[ot]
            measure2 += len(flows2.get(flow, []))
        diff = abs(measure2 - measure1)
        Delta_flow[flow] = diff
        delta_flow[flow] = 1 if diff > theta_flow else 0

    # Fitness Calculation
    N = alpha * len(all_activities) + beta * len(all_flows) + \
        gamma * len(all_activities) + delta * len(all_flows)

    # Calculate numerator components
    fitness_numerator = (alpha *
                         len(A_missing) +
                         beta *
                         len(F_missing) +
                         gamma *
                         sum(delta_act.values()) +
                         delta *
                         sum(delta_flow.values()))

    # To avoid division by zero
    if N == 0:
        fitness = 1.0
    else:
        fitness = 1 - (fitness_numerator / N)
        fitness = max(0.0, min(fitness, 1.0))  # Ensure fitness is within [0,1]

    # Prepare the result dictionary
    result = {
        'missing_activities': A_missing,
        'additional_activities': A_additional,
        'missing_flows': F_missing,
        'additional_flows': F_additional,
        'activity_measure_differences': Delta_act,
        'non_conforming_activities_in_measure': {
            a for a in Delta_act if Delta_act[a] > theta_act},
        'flow_measure_differences': Delta_flow,
        'non_conforming_flows_in_measure': {
            f for f in Delta_flow if Delta_flow[f] > theta_flow},
        'fitness': fitness}

    return result



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any, List, Union
from pm4py.util import exec_utils, constants, pandas_utils
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.algo.transformation.log_to_features import algorithm as log_to_features
from enum import Enum
import numpy as np


class Parameters(Enum):
    INCLUDE_HEADER = "include_header"
    MAX_LEN = "max_len"


def __transform_to_string(stru: str) -> str:
    if stru == "@@max_concurrent_activities_general":
        return "Maximum Number of Concurrent Events"
    elif stru.startswith("@@max_concurrent_activities_like_"):
        return (
            "Maximum Number of Concurrent '"
            + stru.split("@@max_concurrent_activities_like_")[-1]
            + "'"
        )
    elif stru.startswith("event:"):
        stru = stru.split("event:")[-1]
        if "@" in stru:
            attr = stru.split("@")[0]
            value = stru.split("@")[-1]
            return "Value '" + value + "' for Event Attribute '" + attr + "'"
        else:
            return "Values for Event Attribute '" + stru + "'"
    elif stru.startswith("trace:"):
        stru = stru.split("trace:")[-1]
        if "@" in stru:
            attr = stru.split("@")[0]
            value = stru.split("@")[-1]
            return "Value '" + value + "' for Case Attribute '" + attr + "'"
        else:
            return "Values for Case Attribute '" + stru + "'"
    elif stru.startswith("succession:"):
        stru = stru.split("succession:")[-1]
        attr = stru.split("@")[0]
        stru = stru.split("@")[-1]
        val1 = stru.split("#")[0]
        val2 = stru.split("#")[-1]
        return (
            "Succession '"
            + val1
            + "' -> '"
            + val2
            + "' for the Values of the Attribute '"
            + attr
            + "'"
        )
    elif stru == "@@caseDuration":
        return "Case Duration"
    elif stru.startswith("firstIndexAct@@"):
        return (
            "First Position of the Activity '"
            + stru.split("@@")[-1]
            + "' in the Case"
        )
    elif stru.startswith("lastIndexAct@@"):
        return (
            "Last Position of the Activity '"
            + stru.split("@@")[-1]
            + "' in the Case"
        )
    elif stru.startswith("startToLastOcc@@"):
        return (
            "Time from Case Start to Last Occurrence of the Activity '"
            + stru.split("@@")[-1]
            + "'"
        )
    elif stru.startswith("lastOccToEnd@@"):
        return (
            "Time from Last Occurrence of the Activity '"
            + stru.split("@@")[-1]
            + "' to Case End"
        )
    elif stru.startswith("startToFirstOcc@@"):
        return (
            "Time from Case Start to First Occurrence of the Activity '"
            + stru.split("@@")[-1]
            + "'"
        )
    elif stru.startswith("firstOccToEnd@@"):
        return (
            "Time from First Occurrence of the Activity '"
            + stru.split("@@")[-1]
            + "' to Case End"
        )
    elif stru.startswith("directPathPerformanceLastOcc@@"):
        stru = stru.split("@@")[-1].split("##")
        return (
            "Directly-Follows Paths Throughput between '"
            + stru[0]
            + "' and '"
            + stru[1]
            + "' (last occurrence of the path in the case)"
        )
    elif stru.startswith("indirectPathPerformanceLastOcc@@"):
        stru = stru.split("@@")[-1].split("##")
        return (
            "Eventually-Follows Paths Throughput between '"
            + stru[0]
            + "' and '"
            + stru[1]
            + "' (last occurrence of the path in the case)"
        )
    elif stru.startswith("resource_workload@@"):
        return "Resource Workload of '" + stru.split("@@")[-1] + "'"
    elif stru == "@@work_in_progress":
        return "Work in Progress"

    return stru


def _categorize_features(
    fea_df: pd.DataFrame,
) -> List[Dict[str, Any]]:
    """
    Scans the feature table once and returns a list of dictionaries, each of which
    fully describes one *relevant* feature.

    Keys of every dictionary:
        - col (str)  : original column name
        - desc (str) : human label (via __transform_to_string)
        - n_non_zero : how many cases have a non‑zero value
        - std_rel    : relative std‑dev w.r.t. mean (for ranking)
        - series     : the non‑zero part of the column (pd.Series)
    """
    categories: List[Dict[str, Any]] = []

    for col in fea_df.columns:
        ser = fea_df[col]
        non_zero = ser[ser > 0]
        if non_zero.empty:  # skip silent columns
            continue

        mean_val = np.average(non_zero)
        std_rel = 0 if mean_val == 0 or len(non_zero) == 1 else np.std(non_zero) / mean_val
        categories.append(
            dict(
                col=col,
                desc=__transform_to_string(col),
                n_non_zero=len(non_zero),
                std_rel=std_rel,
                series=non_zero,
            )
        )

    # Sort exactly like before: many occurrences first, then variability, then name
    categories.sort(key=lambda d: (d["n_non_zero"], d["std_rel"], d["desc"]), reverse=True)
    return categories


def _features_to_text(
    categories: List[Dict[str, Any]],
    include_header: bool,
    max_len: int,
) -> str:
    parts = ["\n"]
    if include_header:
        parts.append("Given the following features:\n\n")
    text = " ".join(parts)

    for feat in categories:
        if len(text) >= max_len:
            break
        q = feat["series"].quantile([0.0, 0.25, 0.5, 0.75, 1.0]).to_dict()
        block = (
            f"{feat['desc']}:    "
            f"number of non-zero values: {feat['n_non_zero']} ; "
            f"quantiles of the non-zero: {q}\n"
        )
        text += block
    return text


def _features_to_dct(categories: List[Dict[str, Any]]) -> Dict[str, Dict[str, Any]]:
    dct = {}

    for feat in categories:
        q = feat['series'].quantile([0.0, 0.25, 0.5, 0.75, 1.0]).to_dict()
        dct[feat['col']] = {'non_zero': feat['n_non_zero'], 'quantiles': q}

    return dct


def textual_abstraction_from_fea_df(
    fea_df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None
) -> str:
    if parameters is None:
        parameters = {}

    include_header = exec_utils.get_param_value(Parameters.INCLUDE_HEADER, parameters, True)
    max_len = exec_utils.get_param_value(Parameters.MAX_LEN, parameters, constants.OPENAI_MAX_LEN)

    # –– STEP 1: categorise ––––––––––––––––––––––––––––––––––––––––––––––––––––––––––
    categories = _categorize_features(fea_df)

    # –– STEP 2: text rendering ––––––––––––––––––––––––––––––––––––––––––––––––––––––
    return _features_to_text(categories, include_header, max_len)


def dct_abstraction_from_fea_df(fea_df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None) -> Dict[str, Dict[str, Any]]:
    if parameters is None:
        parameters = {}

    categories = _categorize_features(fea_df)

    return _features_to_dct(categories)


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    parameters: Optional[Dict[Any, Any]] = None,
) -> str:
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    data, feature_names = log_to_features.apply(log, parameters=parameters)
    fea_df = pandas_utils.instantiate_dataframe(data, columns=feature_names)

    return textual_abstraction_from_fea_df(fea_df, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
__doc__ = """
The `pm4py.vis` module contains the visualizations offered in pm4py.

**Note on Graphviz-based visualizations (e.g., Petri nets, DFGs, BPMN, process trees):**

Supported output formats include:

- **png**: Generates a PNG image.
- **svg**: Generates an SVG image, which can be scaled without loss of quality.
- **pdf**: Generates a PDF file, suitable for printing and embedding in documents.
- **gv**: Returns or saves the Dot source code of the Graphviz graph.
- **html**: When 'html' is provided as the format, the visualization is rendered in an HTML page using GraphvizJS. This allows interactive viewing of the graph directly in a web browser.

If `html` is used, an HTML file containing the GraphvizJS-based rendering is produced, enabling panning, zooming, and other interactive features.

In general, for Graphviz-based visualizations, if you provide a format extension, the visualization will be generated accordingly. For example:
- `format='png'` will produce a PNG file.
- `format='svg'` will produce an SVG file.
- `format='pdf'` will produce a PDF file.
- `format='gv'` will return/save the raw Dot code.
- `format='html'` will produce an interactive HTML visualization using GraphvizJS.
"""

import os
import sys
from typing import Optional, Union, List, Dict, Any, Tuple, Set

import pandas as pd

from pm4py.objects.bpmn.obj import BPMN
from pm4py.objects.powl.obj import POWL
from pm4py.objects.heuristics_net.obj import HeuristicsNet
from pm4py.objects.log.obj import EventLog
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.util.pandas_utils import (
    check_is_pandas_dataframe,
    check_pandas_dataframe_columns,
)
from pm4py.utils import get_properties
from pm4py.objects.transition_system.obj import TransitionSystem
from pm4py.objects.trie.obj import Trie
from pm4py.objects.ocel.obj import OCEL
from pm4py.objects.org.sna.obj import SNA
from pm4py.util import constants


def _extract_format(format_or_path: str) -> str:
    if "." in format_or_path:
        return os.path.splitext(format_or_path)[1][1:].lower()
    return str(format_or_path).lower()


def _setup_parameters(
    fmt: str,
    bgcolor: str = "white",
    rankdir: str = None,
    graph_title: Optional[str] = None,
) -> Dict[str, Any]:
    parameters = {
        "format": fmt,
        "bgcolor": bgcolor,
        "enable_graph_title": constants.DEFAULT_ENABLE_GRAPH_TITLES,
    }
    if rankdir is not None:
        parameters["rankdir"] = rankdir
        parameters["set_rankdir"] = rankdir
    if graph_title:
        parameters["enable_graph_title"] = True
        parameters["graph_title"] = graph_title
    return parameters


def _select_petri_net_variant(
    variant_str: str,
    pn_visualizer,
    log: Optional[Union[EventLog, pd.DataFrame]],
) -> Any:
    variants_map = {
        "wo_decoration": pn_visualizer.Variants.WO_DECORATION,
        "token_decoration_frequency": pn_visualizer.Variants.FREQUENCY,
        "token_decoration_performance": pn_visualizer.Variants.PERFORMANCE,
        "greedy_decoration_frequency": pn_visualizer.Variants.FREQUENCY_GREEDY,
        "greedy_decoration_performance": pn_visualizer.Variants.PERFORMANCE_GREEDY,
        "alignments": pn_visualizer.Variants.ALIGNMENTS,
    }
    variant = variants_map.get(
        variant_str, pn_visualizer.Variants.WO_DECORATION
    )
    if variant_str != "wo_decoration" and log is None:
        raise Exception(
            "The 'log' parameter must be provided for decoration purposes."
        )
    return variant


def view_petri_net(
    petri_net: PetriNet,
    initial_marking: Optional[Marking] = None,
    final_marking: Optional[Marking] = None,
    format: str = constants.DEFAULT_FORMAT_GVIZ_VIEW,
    bgcolor: str = "white",
    decorations: Dict[Any, Any] = None,
    debug: bool = False,
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    graph_title: Optional[str] = None,
    variant_str: str = "wo_decoration",
    log: Optional[Union[EventLog, pd.DataFrame]] = None,
):
    """
    Views a (composite) Petri net.

    :param petri_net: Petri net
    :param initial_marking: Initial marking
    :param final_marking: Final marking
    :param format: Format of the output picture (if 'html' is provided, GraphvizJS is used to render the visualization in an HTML page)
    :param bgcolor: Background color of the visualization (default: white)
    :param decorations: Decorations (color, label) associated with the elements of the Petri net
    :param debug: Boolean enabling/disabling debug mode (shows place and transition names)
    :param rankdir: Sets the direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)
    :param graph_title: Sets the title of the visualization (if provided)
    :param variant_str: The variant to be used (possible values:
        'wo_decoration', 'token_decoration_frequency', 'token_decoration_performance',
        'greedy_decoration_frequency', 'greedy_decoration_performance', 'alignments')
    :param log: The event log or Pandas dataframe that should be used, if decoration is required

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.discover_petri_net_inductive(dataframe, activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
        pm4py.view_petri_net(net, im, fm, format='svg')
    """
    from pm4py.visualization.petri_net import visualizer as pn_visualizer

    fmt = _extract_format(format)
    variant = _select_petri_net_variant(variant_str, pn_visualizer, log)
    parameters = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    parameters["decorations"] = decorations
    parameters["debug"] = debug
    gviz = pn_visualizer.apply(
        petri_net,
        initial_marking,
        final_marking,
        log=log,
        variant=variant,
        parameters=parameters,
    )
    pn_visualizer.view(gviz)


def save_vis_petri_net(
    petri_net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking,
    file_path: str,
    bgcolor: str = "white",
    decorations: Dict[Any, Any] = None,
    debug: bool = False,
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    graph_title: Optional[str] = None,
    variant_str: str = "wo_decoration",
    log: Optional[Union[EventLog, pd.DataFrame]] = None,
    **kwargs
):
    """
    Saves a Petri net visualization to a file.

    :param petri_net: Petri net
    :param initial_marking: Initial marking
    :param final_marking: Final marking
    :param file_path: Destination path
    :param bgcolor: Background color of the visualization (default: white)
    :param decorations: Decorations (color, label) associated with the elements of the Petri net
    :param debug: Boolean enabling/disabling debug mode (shows place and transition names)
    :param rankdir: Sets the direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)
    :param graph_title: Sets the title of the visualization (if provided)
    :param variant_str: The variant to be used (possible values:
        'wo_decoration', 'token_decoration_frequency', 'token_decoration_performance',
        'greedy_decoration_frequency', 'greedy_decoration_performance', 'alignments')
    :param log: The event log or Pandas dataframe that should be used, if decoration is required

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.discover_petri_net_inductive(dataframe, activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
        pm4py.save_vis_petri_net(net, im, fm, 'petri_net.png')
    """
    from pm4py.visualization.petri_net import visualizer as pn_visualizer

    fmt = _extract_format(file_path)
    variant = _select_petri_net_variant(variant_str, pn_visualizer, log)
    parameters = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    parameters["decorations"] = decorations
    parameters["debug"] = debug
    gviz = pn_visualizer.apply(
        petri_net,
        initial_marking,
        final_marking,
        log=log,
        variant=variant,
        parameters=parameters,
    )
    return pn_visualizer.save(gviz, file_path)


def view_performance_dfg(
    dfg: dict,
    start_activities: dict,
    end_activities: dict,
    format: str = constants.DEFAULT_FORMAT_GVIZ_VIEW,
    aggregation_measure="mean",
    bgcolor: str = "white",
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    serv_time: Optional[Dict[str, float]] = None,
    graph_title: Optional[str] = None,
):
    """
    Views a performance DFG.

    :param dfg: DFG object
    :param start_activities: Start activities
    :param end_activities: End activities
    :param format: Format of the output picture (if 'html' is provided, GraphvizJS is used to render the visualization in an HTML page)
    :param aggregation_measure: Aggregation measure (default: mean), possible values: mean, median, min, max, sum, stdev
    :param bgcolor: Background color of the visualization (default: white)
    :param rankdir: Sets the direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)
    :param serv_time: (optional) Provides the activities' service times, used to decorate the graph
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        performance_dfg, start_activities, end_activities = pm4py.discover_performance_dfg(dataframe, case_id_key='case:concept:name', activity_key='concept:name', timestamp_key='time:timestamp')
        pm4py.view_performance_dfg(performance_dfg, start_activities, end_activities, format='svg')
    """
    fmt = _extract_format(format)
    from pm4py.visualization.dfg import visualizer as dfg_visualizer
    from pm4py.visualization.dfg.variants import (
        performance as dfg_perf_visualizer,
    )

    parameters = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    dfg_parameters = dfg_perf_visualizer.Parameters
    parameters[dfg_parameters.START_ACTIVITIES] = start_activities
    parameters[dfg_parameters.END_ACTIVITIES] = end_activities
    parameters[dfg_parameters.AGGREGATION_MEASURE] = aggregation_measure
    gviz = dfg_perf_visualizer.apply(
        dfg, serv_time=serv_time, parameters=parameters
    )
    dfg_visualizer.view(gviz)


def save_vis_performance_dfg(
    dfg: dict,
    start_activities: dict,
    end_activities: dict,
    file_path: str,
    aggregation_measure="mean",
    bgcolor: str = "white",
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    serv_time: Optional[Dict[str, float]] = None,
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves the visualization of a performance DFG.

    :param dfg: DFG object
    :param start_activities: Start activities
    :param end_activities: End activities
    :param file_path: Destination path
    :param aggregation_measure: Aggregation measure (default: mean), possible values: mean, median, min, max, sum, stdev
    :param bgcolor: Background color of the visualization (default: white)
    :param rankdir: Sets the direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)
    :param serv_time: (optional) Provides the activities' service times, used to decorate the graph
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        performance_dfg, start_activities, end_activities = pm4py.discover_performance_dfg(dataframe, case_id_key='case:concept:name', activity_key='concept:name', timestamp_key='time:timestamp')
        pm4py.save_vis_performance_dfg(performance_dfg, start_activities, end_activities, 'perf_dfg.png')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.dfg import visualizer as dfg_visualizer
    from pm4py.visualization.dfg.variants import (
        performance as dfg_perf_visualizer,
    )

    parameters = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    dfg_parameters = dfg_perf_visualizer.Parameters
    parameters[dfg_parameters.START_ACTIVITIES] = start_activities
    parameters[dfg_parameters.END_ACTIVITIES] = end_activities
    parameters[dfg_parameters.AGGREGATION_MEASURE] = aggregation_measure
    gviz = dfg_perf_visualizer.apply(
        dfg, serv_time=serv_time, parameters=parameters
    )
    return dfg_visualizer.save(gviz, file_path)


def view_dfg(
    dfg: dict,
    start_activities: dict,
    end_activities: dict,
    format: str = constants.DEFAULT_FORMAT_GVIZ_VIEW,
    bgcolor: str = "white",
    max_num_edges: int = sys.maxsize,
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    graph_title: Optional[str] = None,
):
    """
    Views a (composite) DFG.

    :param dfg: DFG object
    :param start_activities: Start activities
    :param end_activities: End activities
    :param format: Format of the output picture (if 'html' is provided, GraphvizJS is used to render the visualization in an HTML page)
    :param bgcolor: Background color of the visualization (default: white)
    :param max_num_edges: Maximum number of edges to represent in the graph
    :param rankdir: Sets the direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        dfg, start_activities, end_activities = pm4py.discover_dfg(dataframe, case_id_key='case:concept:name', activity_key='concept:name', timestamp_key='time:timestamp')
        pm4py.view_dfg(dfg, start_activities, end_activities, format='svg')
    """
    fmt = _extract_format(format)
    from pm4py.visualization.dfg import visualizer as dfg_visualizer

    dfg_parameters = dfg_visualizer.Variants.FREQUENCY.value.Parameters
    parameters = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    parameters["maxNoOfEdgesInDiagram"] = max_num_edges
    parameters[dfg_parameters.START_ACTIVITIES] = start_activities
    parameters[dfg_parameters.END_ACTIVITIES] = end_activities
    gviz = dfg_visualizer.apply(
        dfg, variant=dfg_visualizer.Variants.FREQUENCY, parameters=parameters
    )
    dfg_visualizer.view(gviz)


def save_vis_dfg(
    dfg: dict,
    start_activities: dict,
    end_activities: dict,
    file_path: str,
    bgcolor: str = "white",
    max_num_edges: int = sys.maxsize,
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves a DFG visualization to a file.

    :param dfg: DFG object
    :param start_activities: Start activities
    :param end_activities: End activities
    :param file_path: Destination path
    :param bgcolor: Background color of the visualization (default: white)
    :param max_num_edges: Maximum number of edges to represent in the graph
    :param rankdir: Sets the direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        dfg, start_activities, end_activities = pm4py.discover_dfg(dataframe, case_id_key='case:concept:name', activity_key='concept:name', timestamp_key='time:timestamp')
        pm4py.save_vis_dfg(dfg, start_activities, end_activities, 'dfg.png')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.dfg import visualizer as dfg_visualizer

    dfg_parameters = dfg_visualizer.Variants.FREQUENCY.value.Parameters
    parameters = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    parameters["maxNoOfEdgesInDiagram"] = max_num_edges
    parameters[dfg_parameters.START_ACTIVITIES] = start_activities
    parameters[dfg_parameters.END_ACTIVITIES] = end_activities
    gviz = dfg_visualizer.apply(
        dfg, variant=dfg_visualizer.Variants.FREQUENCY, parameters=parameters
    )
    return dfg_visualizer.save(gviz, file_path)


def view_process_tree(
    tree: ProcessTree,
    format: str = constants.DEFAULT_FORMAT_GVIZ_VIEW,
    bgcolor: str = "white",
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    graph_title: Optional[str] = None,
):
    """
    Views a process tree.

    :param tree: Process tree
    :param format: Format of the visualization (if 'html' is provided, GraphvizJS is used to render the visualization in an HTML page)
    :param bgcolor: Background color of the visualization (default: white)
    :param rankdir: Sets the direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        process_tree = pm4py.discover_process_tree_inductive(dataframe, activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
        pm4py.view_process_tree(process_tree, format='svg')
    """
    fmt = _extract_format(format)
    from pm4py.visualization.process_tree import visualizer as pt_visualizer

    parameters = pt_visualizer.Variants.WO_DECORATION.value.Parameters
    props = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    gviz = pt_visualizer.apply(
        tree, parameters={**props, **{parameters.FORMAT: fmt}}
    )
    pt_visualizer.view(gviz)


def save_vis_process_tree(
    tree: ProcessTree,
    file_path: str,
    bgcolor: str = "white",
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves the visualization of a process tree.

    :param tree: Process tree
    :param file_path: Destination path
    :param bgcolor: Background color of the visualization (default: white)
    :param rankdir: Sets the direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        process_tree = pm4py.discover_process_tree_inductive(dataframe, activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
        pm4py.save_vis_process_tree(process_tree, 'process_tree.png')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.process_tree import visualizer as pt_visualizer

    parameters = pt_visualizer.Variants.WO_DECORATION.value.Parameters
    props = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    gviz = pt_visualizer.apply(
        tree, parameters={**props, **{parameters.FORMAT: fmt}}
    )
    return pt_visualizer.save(gviz, file_path)


def save_vis_bpmn(
    bpmn_graph: BPMN,
    file_path: str,
    bgcolor: str = "white",
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    variant_str: str = "classic",
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves the visualization of a BPMN graph.

    :param bpmn_graph: BPMN graph
    :param file_path: Destination path
    :param bgcolor: Background color of the visualization (default: white)
    :param rankdir: Sets the direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)
    :param variant_str: Variant of the visualization to be used ("classic" or "dagrejs" or "bpmnio_auto_layout")
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        bpmn_graph = pm4py.discover_bpmn_inductive(dataframe, activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
        pm4py.save_vis_bpmn(bpmn_graph, 'trial.bpmn')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.bpmn import visualizer as bpmn_visualizer

    variant = None
    if variant_str == "classic":
        variant = bpmn_visualizer.Variants.CLASSIC
    elif variant_str == "dagrejs":
        variant = bpmn_visualizer.Variants.DAGREJS
    elif variant_str == "bpmnio_auto_layout":
        variant = bpmn_visualizer.Variants.BPMNIO_AUTO_LAYOUT

    props = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    gviz = bpmn_visualizer.apply(bpmn_graph, variant=variant, parameters=props)
    return bpmn_visualizer.save(gviz, file_path, variant=variant)


def view_bpmn(
    bpmn_graph: BPMN,
    format: str = constants.DEFAULT_FORMAT_GVIZ_VIEW,
    bgcolor: str = "white",
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    variant_str: str = "classic",
    graph_title: Optional[str] = None,
):
    """
    Views a BPMN graph.

    :param bpmn_graph: BPMN graph
    :param format: Format of the visualization (if 'html' is provided, GraphvizJS is used to render the visualization in an HTML page)
    :param bgcolor: Background color of the visualization (default: white)
    :param rankdir: Sets the direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)
    :param variant_str: Variant of the visualization to be used ("classic" or "dagrejs" or "bpmnio_auto_layout")
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        bpmn_graph = pm4py.discover_bpmn_inductive(dataframe, activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
        pm4py.view_bpmn(bpmn_graph)
    """
    fmt = _extract_format(format)
    from pm4py.visualization.bpmn import visualizer as bpmn_visualizer

    variant = None
    if variant_str == "classic":
        variant = bpmn_visualizer.Variants.CLASSIC
    elif variant_str == "dagrejs":
        variant = bpmn_visualizer.Variants.DAGREJS
    elif variant_str == "bpmnio_auto_layout":
        variant = bpmn_visualizer.Variants.BPMNIO_AUTO_LAYOUT

    props = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    gviz = bpmn_visualizer.apply(bpmn_graph, variant=variant, parameters=props)
    bpmn_visualizer.view(gviz, variant=variant)


def view_heuristics_net(
    heu_net: HeuristicsNet,
    format: str = "png",
    bgcolor: str = "white",
    graph_title: Optional[str] = None,
):
    """
    Views a heuristics net.

    :param heu_net: Heuristics net
    :param format: Format of the visualization
    :param bgcolor: Background color of the visualization (default: white)
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        heu_net = pm4py.discover_heuristics_net(dataframe, activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
        pm4py.view_heuristics_net(heu_net, format='svg')
    """
    fmt = _extract_format(format)
    from pm4py.visualization.heuristics_net import visualizer as hn_visualizer

    parameters = hn_visualizer.Variants.PYDOTPLUS.value.Parameters
    props = _setup_parameters(fmt, bgcolor, graph_title=graph_title)
    gviz = hn_visualizer.apply(
        heu_net, parameters={**props, **{parameters.FORMAT: fmt}}
    )
    hn_visualizer.view(gviz)


def save_vis_heuristics_net(
    heu_net: HeuristicsNet,
    file_path: str,
    bgcolor: str = "white",
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves the visualization of a heuristics net.

    :param heu_net: Heuristics net
    :param file_path: Destination path
    :param bgcolor: Background color of the visualization (default: white)
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        heu_net = pm4py.discover_heuristics_net(dataframe, activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
        pm4py.save_vis_heuristics_net(heu_net, 'heu.png')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.heuristics_net import visualizer as hn_visualizer

    parameters = hn_visualizer.Variants.PYDOTPLUS.value.Parameters
    props = _setup_parameters(fmt, bgcolor, graph_title=graph_title)
    gviz = hn_visualizer.apply(
        heu_net, parameters={**props, **{parameters.FORMAT: fmt}}
    )
    return hn_visualizer.save(gviz, file_path)


def __dotted_attribute_selection(
    log: Union[EventLog, pd.DataFrame], attributes
):
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log)

    if attributes is None:
        if isinstance(log, EventLog):
            for index, trace in enumerate(log):
                trace.attributes["@@index"] = index
            attributes = ["time:timestamp", "case:@@index", "concept:name"]
        else:
            attributes = ["time:timestamp", "@@case_index", "concept:name"]

    return log, attributes


def view_dotted_chart(
    log: Union[EventLog, pd.DataFrame],
    format: str = "png",
    attributes=None,
    bgcolor: str = "white",
    show_legend: bool = True,
    graph_title: Optional[str] = None,
):
    """
    Displays the dotted chart.

    Each event in the log is represented as a point. Dimensions are:
    - X-axis: The value of the first selected attribute.
    - Y-axis: The value of the second selected attribute.
    - Color: The value of the third selected attribute.

    If attributes are not provided, a default dotted chart is shown:
    X-axis: time
    Y-axis: case index (in order of occurrence)
    Color: activity

    :param log: Event log
    :param format: Image format
    :param attributes: Attributes used to construct the dotted chart. If None, the default dotted chart is used. For custom attributes, use a list of the form [x-axis attribute, y-axis attribute, color attribute].
    :param bgcolor: Background color of the chart (default: white)
    :param show_legend: Boolean (enables/disables the legend)
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        pm4py.view_dotted_chart(dataframe, format='svg')
        pm4py.view_dotted_chart(dataframe, attributes=['time:timestamp', 'concept:name', 'org:resource'])
    """
    fmt = _extract_format(format)
    log, attributes = __dotted_attribute_selection(log, attributes)
    from pm4py.visualization.dotted_chart import (
        visualizer as dotted_chart_visualizer,
    )

    params = _setup_parameters(fmt, bgcolor, graph_title=graph_title)
    params["show_legend"] = show_legend
    gviz = dotted_chart_visualizer.apply(log, attributes, parameters=params)
    dotted_chart_visualizer.view(gviz)


def save_vis_dotted_chart(
    log: Union[EventLog, pd.DataFrame],
    file_path: str,
    attributes=None,
    bgcolor: str = "white",
    show_legend: bool = True,
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves the visualization of the dotted chart.

    Each event in the log is represented as a point. Dimensions are:
    - X-axis: The value of the first selected attribute.
    - Y-axis: The value of the second selected attribute.
    - Color: The value of the third selected attribute.

    If attributes are not provided, a default dotted chart is used:
    X-axis: time
    Y-axis: case index (in order of occurrence)
    Color: activity

    :param log: Event log
    :param file_path: Destination path
    :param attributes: Attributes for the dotted chart. For example, ["time:timestamp", "concept:name", "org:resource"].
    :param bgcolor: Background color of the chart (default: white)
    :param show_legend: Boolean (enables/disables the legend)
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        pm4py.save_vis_dotted_chart(dataframe, 'dotted.png', attributes=['time:timestamp', 'concept:name', 'org:resource'])
    """
    fmt = _extract_format(file_path)
    log, attributes = __dotted_attribute_selection(log, attributes)
    from pm4py.visualization.dotted_chart import (
        visualizer as dotted_chart_visualizer,
    )

    params = _setup_parameters(fmt, bgcolor, graph_title=graph_title)
    params["show_legend"] = show_legend
    gviz = dotted_chart_visualizer.apply(log, attributes, parameters=params)
    return dotted_chart_visualizer.save(gviz, file_path)


def view_sna(sna_metric: SNA, variant_str: Optional[str] = None):
    """
    Represents a SNA metric (.html).

    :param sna_metric: Values of the metric
    :param variant_str: Variant to be used (default: pyvis)

    .. code-block:: python3

        import pm4py

        metric = pm4py.discover_subcontracting_network(dataframe, resource_key='org:resource', timestamp_key='time:timestamp', case_id_key='case:concept:name')
        pm4py.view_sna(metric)
    """
    if variant_str is None:
        if constants.DEFAULT_GVIZ_VIEW == "matplotlib_view":
            variant_str = "networkx"
        else:
            variant_str = "pyvis"

    from pm4py.visualization.sna import visualizer as sna_visualizer

    variant = sna_visualizer.Variants.PYVIS
    if variant_str == "networkx":
        variant = sna_visualizer.Variants.NETWORKX
    gviz = sna_visualizer.apply(sna_metric, variant=variant)
    sna_visualizer.view(gviz, variant=variant)


def save_vis_sna(
    sna_metric: SNA,
    file_path: str,
    variant_str: Optional[str] = None,
    **kwargs
):
    """
    Saves the visualization of a SNA metric in a .html file.

    :param sna_metric: Values of the metric
    :param file_path: Destination path
    :param variant_str: Variant to be used (default: pyvis)

    .. code-block:: python3

        import pm4py

        metric = pm4py.discover_subcontracting_network(dataframe, resource_key='org:resource', timestamp_key='time:timestamp', case_id_key='case:concept:name')
        pm4py.save_vis_sna(metric, 'sna.png')
    """
    if variant_str is None:
        if constants.DEFAULT_GVIZ_VIEW == "matplotlib_view":
            variant_str = "networkx"
        else:
            variant_str = "pyvis"

    from pm4py.visualization.sna import visualizer as sna_visualizer

    variant = sna_visualizer.Variants.PYVIS
    if variant_str == "networkx":
        variant = sna_visualizer.Variants.NETWORKX

    gviz = sna_visualizer.apply(sna_metric, variant=variant)
    return sna_visualizer.save(gviz, file_path, variant=variant)


def view_case_duration_graph(
    log: Union[EventLog, pd.DataFrame],
    format: str = "png",
    activity_key="concept:name",
    timestamp_key="time:timestamp",
    case_id_key="case:concept:name",
    graph_title: Optional[str] = None,
):
    """
    Visualizes the case duration graph.

    :param log: Log object
    :param format: Format of the visualization (png, svg, ...)
    :param activity_key: Attribute to be used as activity
    :param case_id_key: Attribute to be used as case identifier
    :param timestamp_key: Attribute to be used as timestamp
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        pm4py.view_case_duration_graph(dataframe, format='svg', activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
    """
    fmt = _extract_format(format)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log,
            activity_key=activity_key,
            case_id_key=case_id_key,
            timestamp_key=timestamp_key,
        )
        from pm4py.statistics.traces.generic.pandas import case_statistics

        graph = case_statistics.get_kde_caseduration(
            log,
            parameters=get_properties(
                log,
                activity_key=activity_key,
                case_id_key=case_id_key,
                timestamp_key=timestamp_key,
            ),
        )
    else:
        from pm4py.statistics.traces.generic.log import case_statistics

        graph = case_statistics.get_kde_caseduration(
            log,
            parameters=get_properties(
                log,
                activity_key=activity_key,
                case_id_key=case_id_key,
                timestamp_key=timestamp_key,
            ),
        )
    from pm4py.visualization.graphs import visualizer as graphs_visualizer

    properties = {"format": fmt}
    if graph_title is not None:
        properties["title"] = graph_title

    graph_vis = graphs_visualizer.apply(
        graph[0],
        graph[1],
        variant=graphs_visualizer.Variants.CASES,
        parameters=properties,
    )
    graphs_visualizer.view(graph_vis)


def save_vis_case_duration_graph(
    log: Union[EventLog, pd.DataFrame],
    file_path: str,
    activity_key="concept:name",
    timestamp_key="time:timestamp",
    case_id_key="case:concept:name",
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves the case duration graph to the specified path.

    :param log: Log object
    :param file_path: Destination path
    :param activity_key: Attribute to be used as activity
    :param case_id_key: Attribute to be used as case identifier
    :param timestamp_key: Attribute to be used as timestamp
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        pm4py.save_vis_case_duration_graph(dataframe, 'duration.png', activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
    """
    fmt = _extract_format(file_path)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log,
            activity_key=activity_key,
            case_id_key=case_id_key,
            timestamp_key=timestamp_key,
        )
        from pm4py.statistics.traces.generic.pandas import case_statistics

        graph = case_statistics.get_kde_caseduration(
            log,
            parameters=get_properties(
                log,
                activity_key=activity_key,
                case_id_key=case_id_key,
                timestamp_key=timestamp_key,
            ),
        )
    else:
        from pm4py.statistics.traces.generic.log import case_statistics

        graph = case_statistics.get_kde_caseduration(
            log,
            parameters=get_properties(
                log,
                activity_key=activity_key,
                case_id_key=case_id_key,
                timestamp_key=timestamp_key,
            ),
        )
    from pm4py.visualization.graphs import visualizer as graphs_visualizer

    properties = {"format": fmt}
    if graph_title is not None:
        properties["title"] = graph_title

    graph_vis = graphs_visualizer.apply(
        graph[0],
        graph[1],
        variant=graphs_visualizer.Variants.CASES,
        parameters=properties,
    )
    return graphs_visualizer.save(graph_vis, file_path)


def view_events_per_time_graph(
    log: Union[EventLog, pd.DataFrame],
    format: str = "png",
    activity_key="concept:name",
    timestamp_key="time:timestamp",
    case_id_key="case:concept:name",
    graph_title: Optional[str] = None,
):
    """
    Visualizes the events per time graph.

    :param log: Log object
    :param format: Format of the visualization (png, svg, ...)
    :param activity_key: Attribute to be used as activity
    :param case_id_key: Attribute to be used as case identifier
    :param timestamp_key: Attribute to be used as timestamp
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        pm4py.view_events_per_time_graph(dataframe, format='svg', activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
    """
    fmt = _extract_format(format)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log,
            activity_key=activity_key,
            case_id_key=case_id_key,
            timestamp_key=timestamp_key,
        )
        from pm4py.statistics.attributes.pandas import get as attributes_get

        graph = attributes_get.get_kde_date_attribute(
            log,
            parameters=get_properties(
                log,
                activity_key=activity_key,
                case_id_key=case_id_key,
                timestamp_key=timestamp_key,
            ),
        )
    else:
        from pm4py.statistics.attributes.log import get as attributes_get

        graph = attributes_get.get_kde_date_attribute(
            log,
            parameters=get_properties(
                log,
                activity_key=activity_key,
                case_id_key=case_id_key,
                timestamp_key=timestamp_key,
            ),
        )
    from pm4py.visualization.graphs import visualizer as graphs_visualizer

    properties = {"format": fmt}
    if graph_title is not None:
        properties["title"] = graph_title

    graph_vis = graphs_visualizer.apply(
        graph[0],
        graph[1],
        variant=graphs_visualizer.Variants.DATES,
        parameters=properties,
    )
    graphs_visualizer.view(graph_vis)


def save_vis_events_per_time_graph(
    log: Union[EventLog, pd.DataFrame],
    file_path: str,
    activity_key="concept:name",
    timestamp_key="time:timestamp",
    case_id_key="case:concept:name",
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves the events per time graph to the specified path.

    :param log: Log object
    :param file_path: Destination path
    :param activity_key: Attribute to be used as activity
    :param case_id_key: Attribute to be used as case identifier
    :param timestamp_key: Attribute to be used as timestamp
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        pm4py.save_vis_events_per_time_graph(dataframe, 'ev_time.png', activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
    """
    fmt = _extract_format(file_path)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log,
            activity_key=activity_key,
            case_id_key=case_id_key,
            timestamp_key=timestamp_key,
        )
        from pm4py.statistics.attributes.pandas import get as attributes_get

        graph = attributes_get.get_kde_date_attribute(
            log,
            attribute=timestamp_key,
            parameters=get_properties(
                log,
                activity_key=activity_key,
                case_id_key=case_id_key,
                timestamp_key=timestamp_key,
            ),
        )
    else:
        from pm4py.statistics.attributes.log import get as attributes_get

        graph = attributes_get.get_kde_date_attribute(
            log,
            attribute=timestamp_key,
            parameters=get_properties(
                log,
                activity_key=activity_key,
                case_id_key=case_id_key,
                timestamp_key=timestamp_key,
            ),
        )
    fmt = _extract_format(file_path)
    from pm4py.visualization.graphs import visualizer as graphs_visualizer

    properties = {"format": fmt}
    if graph_title is not None:
        properties["title"] = graph_title

    graph_vis = graphs_visualizer.apply(
        graph[0],
        graph[1],
        variant=graphs_visualizer.Variants.DATES,
        parameters=properties,
    )

    return graphs_visualizer.save(graph_vis, file_path)


def view_performance_spectrum(
    log: Union[EventLog, pd.DataFrame],
    activities: List[str],
    format: str = "png",
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
    bgcolor: str = "white",
    graph_title: Optional[str] = None,
):
    """
    Displays the performance spectrum.

    The performance spectrum is a novel visualization of the performance of the process, showing time elapsed between different activities.
    Refer to:
    Denisov, Vadim, et al. "The Performance Spectrum Miner: Visual Analytics for Fine-Grained Performance Analysis of Processes."
    BPM (Dissertation/Demos/Industry). 2018.

    :param log: Event log
    :param activities: List of activities (in order) used to build the performance spectrum
    :param format: Format of the visualization (png, svg ...)
    :param activity_key: Attribute to be used for the activity
    :param timestamp_key: Attribute to be used for the timestamp
    :param case_id_key: Attribute to be used as case identifier
    :param bgcolor: Background color of the visualization (default: white)
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        pm4py.view_performance_spectrum(dataframe, ['Act. A', 'Act. C', 'Act. D'], format='svg', activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
    """
    fmt = _extract_format(format)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log,
            activity_key=activity_key,
            case_id_key=case_id_key,
            timestamp_key=timestamp_key,
        )

    properties = get_properties(
        log,
        activity_key=activity_key,
        case_id_key=case_id_key,
        timestamp_key=timestamp_key,
    )
    from pm4py.algo.discovery.performance_spectrum import (
        algorithm as performance_spectrum,
    )

    perf_spectrum = performance_spectrum.apply(
        log, activities, parameters=properties
    )
    from pm4py.visualization.performance_spectrum import (
        visualizer as perf_spectrum_visualizer,
    )
    from pm4py.visualization.performance_spectrum.variants import neato

    params = _setup_parameters(fmt, bgcolor, graph_title=graph_title)
    gviz = perf_spectrum_visualizer.apply(
        perf_spectrum,
        parameters={**params, **{neato.Parameters.FORMAT.value: fmt}},
    )
    perf_spectrum_visualizer.view(gviz)


def save_vis_performance_spectrum(
    log: Union[EventLog, pd.DataFrame],
    activities: List[str],
    file_path: str,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
    bgcolor: str = "white",
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves the visualization of the performance spectrum to a file.

    Refer to:
    Denisov, Vadim, et al. "The Performance Spectrum Miner: Visual Analytics for Fine-Grained Performance Analysis of Processes."
    BPM (Dissertation/Demos/Industry). 2018.

    :param log: Event log
    :param activities: List of activities used to build the performance spectrum
    :param file_path: Destination path (including the extension)
    :param activity_key: Attribute to be used for the activity
    :param timestamp_key: Attribute to be used for the timestamp
    :param case_id_key: Attribute to be used as case identifier
    :param bgcolor: Background color of the visualization (default: white)
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        pm4py.save_vis_performance_spectrum(dataframe, ['Act. A', 'Act. C', 'Act. D'], 'perf_spec.png', activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
    """
    fmt = _extract_format(file_path)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log,
            activity_key=activity_key,
            case_id_key=case_id_key,
            timestamp_key=timestamp_key,
        )

    properties = get_properties(
        log,
        activity_key=activity_key,
        case_id_key=case_id_key,
        timestamp_key=timestamp_key,
    )
    from pm4py.algo.discovery.performance_spectrum import (
        algorithm as performance_spectrum,
    )

    perf_spectrum = performance_spectrum.apply(
        log, activities, parameters=properties
    )
    from pm4py.visualization.performance_spectrum import (
        visualizer as perf_spectrum_visualizer,
    )
    from pm4py.visualization.performance_spectrum.variants import neato

    params = _setup_parameters(fmt, bgcolor, graph_title=graph_title)
    gviz = perf_spectrum_visualizer.apply(
        perf_spectrum,
        parameters={**params, **{neato.Parameters.FORMAT.value: fmt}},
    )
    return perf_spectrum_visualizer.save(gviz, file_path)


def __builds_events_distribution_graph(
    log: Union[EventLog, pd.DataFrame],
    parameters,
    distr_type: str = "days_week",
):
    if distr_type == "days_month":
        title = "Distribution of the Events over the Days of a Month"
        x_axis = "Day of month"
        y_axis = "Number of Events"
    elif distr_type == "months":
        title = "Distribution of the Events over the Months"
        x_axis = "Month"
        y_axis = "Number of Events"
    elif distr_type == "years":
        title = "Distribution of the Events over the Years"
        x_axis = "Year"
        y_axis = "Number of Events"
    elif distr_type == "hours":
        title = "Distribution of the Events over the Hours"
        x_axis = "Hour (of day)"
        y_axis = "Number of Events"
    elif distr_type == "days_week":
        title = "Distribution of the Events over the Days of a Week"
        x_axis = "Day of the Week"
        y_axis = "Number of Events"
    elif distr_type == "weeks":
        title = "Distribution of the Events over the Weeks of a Year"
        x_axis = "Week of the Year"
        y_axis = "Number of Events"
    else:
        raise Exception("Unsupported distribution specified.")

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log)
        from pm4py.statistics.attributes.pandas import get as attributes_get

        x, y = attributes_get.get_events_distribution(
            log, distr_type=distr_type, parameters=parameters
        )
    else:
        from pm4py.statistics.attributes.log import get as attributes_get

        x, y = attributes_get.get_events_distribution(
            log, distr_type=distr_type, parameters=parameters
        )

    return title, x_axis, y_axis, x, y


def view_events_distribution_graph(
    log: Union[EventLog, pd.DataFrame],
    distr_type: str = "days_week",
    format="png",
    activity_key="concept:name",
    timestamp_key="time:timestamp",
    case_id_key="case:concept:name",
    graph_title: Optional[str] = None,
):
    """
    Shows the distribution of the events in the specified dimension.

    This allows identifying work shifts, busy days, and busy periods of the year.

    :param log: Event log
    :param distr_type: Type of distribution (default: days_week):
        - days_month: Distribution of events among days of a month (1-31)
        - months: Distribution of events among months (1-12)
        - years: Distribution of events among years
        - hours: Distribution of events among hours of a day (0-23)
        - days_week: Distribution of events among days of the week (Mon-Sun)
        - weeks: Distribution of events among weeks of a year (0-52)
    :param format: Format of the visualization (default: png)
    :param activity_key: Attribute to be used as activity
    :param case_id_key: Attribute to be used as case identifier
    :param timestamp_key: Attribute to be used as timestamp
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        pm4py.view_events_distribution_graph(dataframe, format='svg', distr_type='days_week', activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
    """
    fmt = _extract_format(format)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log,
            activity_key=activity_key,
            case_id_key=case_id_key,
            timestamp_key=timestamp_key,
        )

    parameters = get_properties(
        log,
        activity_key=activity_key,
        case_id_key=case_id_key,
        timestamp_key=timestamp_key,
    )
    title, x_axis, y_axis, x, y = __builds_events_distribution_graph(
        log, parameters, distr_type
    )
    parameters["title"] = graph_title if graph_title else title
    parameters["x_axis"] = x_axis
    parameters["y_axis"] = y_axis
    parameters["format"] = fmt

    from pm4py.visualization.graphs import visualizer as graphs_visualizer

    gviz = graphs_visualizer.apply(
        x, y, variant=graphs_visualizer.Variants.BARPLOT, parameters=parameters
    )
    graphs_visualizer.view(gviz)


def save_vis_events_distribution_graph(
    log: Union[EventLog, pd.DataFrame],
    file_path: str,
    distr_type: str = "days_week",
    activity_key="concept:name",
    timestamp_key="time:timestamp",
    case_id_key="case:concept:name",
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves the distribution of the events in a picture file.

    Observing the distribution of events over time helps infer work shifts, working days, and busy periods of the year.

    :param log: Event log
    :param file_path: Destination path (including the extension)
    :param distr_type: Type of distribution (default: days_week):
        - days_month: Events distribution among days of a month (1-31)
        - months: Events distribution among months (1-12)
        - years: Events distribution among years
        - hours: Events distribution among hours of a day (0-23)
        - days_week: Events distribution among days of a week (Mon-Sun)
        - weeks: Events distribution among weeks of a year (0-52)
    :param activity_key: Attribute to be used as activity
    :param case_id_key: Attribute to be used as case identifier
    :param timestamp_key: Attribute to be used as timestamp
    :param graph_title: Sets the title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        pm4py.save_vis_events_distribution_graph(dataframe, 'ev_distr_graph.png', distr_type='days_week', activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
    """
    fmt = _extract_format(file_path)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log,
            activity_key=activity_key,
            case_id_key=case_id_key,
            timestamp_key=timestamp_key,
        )

    parameters = get_properties(
        log,
        activity_key=activity_key,
        case_id_key=case_id_key,
        timestamp_key=timestamp_key,
    )
    title, x_axis, y_axis, x, y = __builds_events_distribution_graph(
        log, parameters, distr_type
    )
    parameters["title"] = graph_title if graph_title else title
    parameters["x_axis"] = x_axis
    parameters["y_axis"] = y_axis
    parameters["format"] = fmt

    from pm4py.visualization.graphs import visualizer as graphs_visualizer

    gviz = graphs_visualizer.apply(
        x, y, variant=graphs_visualizer.Variants.BARPLOT, parameters=parameters
    )
    return graphs_visualizer.save(gviz, file_path)


def view_ocdfg(
    ocdfg: Dict[str, Any],
    annotation: str = "frequency",
    act_metric: str = "events",
    edge_metric="event_couples",
    act_threshold: int = 0,
    edge_threshold: int = 0,
    performance_aggregation: str = "mean",
    format: str = constants.DEFAULT_FORMAT_GVIZ_VIEW,
    bgcolor: str = "white",
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    graph_title: Optional[str] = None,
    variant_str: str = "classic",
):
    """
    Views an OC-DFG (object-centric directly-follows graph).

    :param ocdfg: Object-centric directly-follows graph
    :param annotation: The annotation to use ("frequency" or "performance")
    :param act_metric: The metric for activities ("events", "unique_objects", "total_objects")
    :param edge_metric: The metric for edges ("event_couples", "unique_objects", "total_objects")
    :param act_threshold: Threshold on activities frequency (default: 0)
    :param edge_threshold: Threshold on edges frequency (default: 0)
    :param performance_aggregation: Aggregation measure for performance: mean, median, min, max, sum
    :param format: Format of the output (if 'html' is provided, GraphvizJS is used)
    :param bgcolor: Background color (default: white)
    :param rankdir: Graph direction ("LR" or "TB")
    :param graph_title: Title of the visualization (if provided)
    :param variant_str: Variant of the visualization ("classic" or "elkjs")

    .. code-block:: python3

        import pm4py

        ocdfg = pm4py.discover_ocdfg(ocel)
        pm4py.view_ocdfg(ocdfg, annotation='frequency', format='svg')
    """
    fmt = _extract_format(format)
    from pm4py.visualization.ocel.ocdfg import visualizer

    parameters = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    parameters["annotation"] = annotation
    parameters["act_metric"] = act_metric
    parameters["edge_metric"] = edge_metric
    parameters["act_threshold"] = act_threshold
    parameters["edge_threshold"] = edge_threshold
    parameters["aggregation_measure"] = performance_aggregation

    variant = (
        visualizer.Variants.CLASSIC
        if variant_str == "classic"
        else visualizer.Variants.ELKJS
    )

    gviz = visualizer.apply(ocdfg, variant=variant, parameters=parameters)
    visualizer.view(gviz, variant=variant)


def save_vis_ocdfg(
    ocdfg: Dict[str, Any],
    file_path: str,
    annotation: str = "frequency",
    act_metric: str = "events",
    edge_metric="event_couples",
    act_threshold: int = 0,
    edge_threshold: int = 0,
    performance_aggregation: str = "mean",
    bgcolor: str = "white",
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    graph_title: Optional[str] = None,
    variant_str: str = "classic",
    **kwargs
):
    """
    Saves the visualization of an OC-DFG.

    :param ocdfg: Object-centric directly-follows graph
    :param file_path: Destination path
    :param annotation: "frequency" or "performance"
    :param act_metric: Metric for activities ("events", "unique_objects", "total_objects")
    :param edge_metric: Metric for edges ("event_couples", "unique_objects", "total_objects")
    :param act_threshold: Threshold on activities frequency
    :param edge_threshold: Threshold on edges frequency
    :param performance_aggregation: Aggregation measure for performance: mean, median, min, max, sum
    :param bgcolor: Background color (default: white)
    :param rankdir: Graph direction ("LR" or "TB")
    :param graph_title: Title of the visualization (if provided)
    :param variant_str: Variant ("classic" or "elkjs")

    .. code-block:: python3

        import pm4py

        ocdfg = pm4py.discover_ocdfg(ocel)
        pm4py.save_vis_ocdfg(ocdfg, 'ocdfg.png', annotation='frequency')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.ocel.ocdfg import visualizer

    parameters = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    parameters["annotation"] = annotation
    parameters["act_metric"] = act_metric
    parameters["edge_metric"] = edge_metric
    parameters["act_threshold"] = act_threshold
    parameters["edge_threshold"] = edge_threshold
    parameters["aggregation_measure"] = performance_aggregation

    variant = (
        visualizer.Variants.CLASSIC
        if variant_str == "classic"
        else visualizer.Variants.ELKJS
    )
    gviz = visualizer.apply(ocdfg, variant=variant, parameters=parameters)
    return visualizer.save(gviz, file_path, variant=variant)


def view_ocpn(
    ocpn: Dict[str, Any],
    format: str = constants.DEFAULT_FORMAT_GVIZ_VIEW,
    bgcolor: str = "white",
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    graph_title: Optional[str] = None,
    variant_str: str = "wo_decoration"
):
    """
    Visualizes the object-centric Petri net.

    :param ocpn: Object-centric Petri net
    :param format: Format of the visualization (if 'html' is provided, GraphvizJS is used)
    :param bgcolor: Background color (default: white)
    :param rankdir: Graph direction ("LR" or "TB")
    :param graph_title: Title of the visualization (if provided)
    :param variant_str: Variant to be used ("wo_decoration" or "brachmann")

    .. code-block:: python3

        import pm4py

        ocpn = pm4py.discover_oc_petri_net(ocel)
        pm4py.view_ocpn(ocpn, format='svg')
    """
    fmt = _extract_format(format)
    from pm4py.visualization.ocel.ocpn import visualizer as ocpn_visualizer

    props = _setup_parameters(fmt, bgcolor, rankdir, graph_title)

    variant = (
        ocpn_visualizer.Variants.WO_DECORATION
        if variant_str == "wo_decoration"
        else ocpn_visualizer.Variants.BRACHMANN
    )

    gviz = ocpn_visualizer.apply(ocpn, variant=variant, parameters=props)
    ocpn_visualizer.view(gviz)


def save_vis_ocpn(
    ocpn: Dict[str, Any],
    file_path: str,
    bgcolor: str = "white",
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    graph_title: Optional[str] = None,
    variant_str: str = "wo_decoration",
    **kwargs
):
    """
    Saves the visualization of the object-centric Petri net into a file.

    :param ocpn: Object-centric Petri net
    :param file_path: Target path
    :param bgcolor: Background color (default: white)
    :param rankdir: Graph direction ("LR" or "TB")
    :param graph_title: Title of the visualization (if provided)
    :param variant_str: Variant to be used ("wo_decoration" or "brachmann")

    .. code-block:: python3

        import pm4py

        ocpn = pm4py.discover_oc_petri_net(ocel)
        pm4py.save_vis_ocpn(ocpn, 'ocpn.png')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.ocel.ocpn import visualizer as ocpn_visualizer

    props = _setup_parameters(fmt, bgcolor, rankdir, graph_title)

    variant = (
        ocpn_visualizer.Variants.WO_DECORATION
        if variant_str == "wo_decoration"
        else ocpn_visualizer.Variants.BRACHMANN
    )

    gviz = ocpn_visualizer.apply(ocpn, variant=variant, parameters=props)
    return ocpn_visualizer.save(gviz, file_path)


def view_network_analysis(
    network_analysis: Dict[Tuple[str, str], Dict[str, Any]],
    variant: str = "frequency",
    format: str = constants.DEFAULT_FORMAT_GVIZ_VIEW,
    activity_threshold: int = 1,
    edge_threshold: int = 1,
    bgcolor: str = "white",
    graph_title: Optional[str] = None,
):
    """
    Visualizes the network analysis.

    :param network_analysis: Network analysis
    :param variant: "frequency" or "performance"
    :param format: Format of the visualization (if 'html' is provided, GraphvizJS is used)
    :param activity_threshold: Minimum occurrences of an activity to be included
    :param edge_threshold: Minimum occurrences of an edge to be included
    :param bgcolor: Background color (default: white)
    :param graph_title: Title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        net_ana = pm4py.discover_network_analysis(dataframe, out_column='case:concept:name', in_column='case:concept:name', node_column_source='org:resource', node_column_target='org:resource', edge_column='concept:name')
        pm4py.view_network_analysis(net_ana, format='svg')
    """
    fmt = _extract_format(format)
    from pm4py.visualization.network_analysis import (
        visualizer as network_analysis_visualizer,
    )

    chosen_variant = (
        network_analysis_visualizer.Variants.PERFORMANCE
        if variant == "performance"
        else network_analysis_visualizer.Variants.FREQUENCY
    )
    params = _setup_parameters(fmt, bgcolor, graph_title=graph_title)
    params["activity_threshold"] = activity_threshold
    params["edge_threshold"] = edge_threshold

    gviz = network_analysis_visualizer.apply(
        network_analysis, variant=chosen_variant, parameters=params
    )
    network_analysis_visualizer.view(gviz)


def save_vis_network_analysis(
    network_analysis: Dict[Tuple[str, str], Dict[str, Any]],
    file_path: str,
    variant: str = "frequency",
    activity_threshold: int = 1,
    edge_threshold: int = 1,
    bgcolor: str = "white",
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves the visualization of the network analysis.

    :param network_analysis: Network analysis
    :param file_path: Target path
    :param variant: "frequency" or "performance"
    :param activity_threshold: Minimum occurrences of an activity
    :param edge_threshold: Minimum occurrences of an edge
    :param bgcolor: Background color (default: white)
    :param graph_title: Title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        net_ana = pm4py.discover_network_analysis(dataframe, out_column='case:concept:name', in_column='case:concept:name', node_column_source='org:resource', node_column_target='org:resource', edge_column='concept:name')
        pm4py.save_vis_network_analysis(net_ana, 'net_ana.png')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.network_analysis import (
        visualizer as network_analysis_visualizer,
    )

    chosen_variant = (
        network_analysis_visualizer.Variants.PERFORMANCE
        if variant == "performance"
        else network_analysis_visualizer.Variants.FREQUENCY
    )
    params = _setup_parameters(fmt, bgcolor, graph_title=graph_title)
    params["activity_threshold"] = activity_threshold
    params["edge_threshold"] = edge_threshold

    gviz = network_analysis_visualizer.apply(
        network_analysis, variant=chosen_variant, parameters=params
    )
    return network_analysis_visualizer.save(gviz, file_path)


def view_transition_system(
    transition_system: TransitionSystem,
    format: str = constants.DEFAULT_FORMAT_GVIZ_VIEW,
    bgcolor: str = "white",
    graph_title: Optional[str] = None,
):
    """
    Views a transition system.

    :param transition_system: Transition system
    :param format: Format of the visualization (if 'html' is provided, GraphvizJS is used)
    :param bgcolor: Background color (default: white)
    :param graph_title: Title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        transition_system = pm4py.discover_transition_system(dataframe, activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
        pm4py.view_transition_system(transition_system, format='svg')
    """
    fmt = _extract_format(format)
    from pm4py.visualization.transition_system import (
        visualizer as ts_visualizer,
    )

    params = _setup_parameters(fmt, bgcolor, graph_title=graph_title)
    gviz = ts_visualizer.apply(transition_system, parameters=params)
    ts_visualizer.view(gviz)


def save_vis_transition_system(
    transition_system: TransitionSystem,
    file_path: str,
    bgcolor: str = "white",
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Persists the visualization of a transition system.

    :param transition_system: Transition system
    :param file_path: Destination path
    :param bgcolor: Background color (default: white)
    :param graph_title: Title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        transition_system = pm4py.discover_transition_system(dataframe, activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
        pm4py.save_vis_transition_system(transition_system, 'trans_system.png')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.transition_system import (
        visualizer as ts_visualizer,
    )

    params = _setup_parameters(fmt, bgcolor, graph_title=graph_title)
    gviz = ts_visualizer.apply(transition_system, parameters=params)
    return ts_visualizer.save(gviz, file_path)


def view_prefix_tree(
    trie: Trie,
    format: str = constants.DEFAULT_FORMAT_GVIZ_VIEW,
    bgcolor: str = "white",
    graph_title: Optional[str] = None,
):
    """
    Views a prefix tree.

    :param trie: Prefix tree
    :param format: Format of the visualization (if 'html' is provided, GraphvizJS is used)
    :param bgcolor: Background color (default: white)
    :param graph_title: Title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        prefix_tree = pm4py.discover_prefix_tree(dataframe, activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
        pm4py.view_prefix_tree(prefix_tree, format='svg')
    """
    fmt = _extract_format(format)
    from pm4py.visualization.trie import visualizer as trie_visualizer

    params = _setup_parameters(fmt, bgcolor, graph_title=graph_title)
    gviz = trie_visualizer.apply(trie, parameters=params)
    trie_visualizer.view(gviz)


def save_vis_prefix_tree(
    trie: Trie,
    file_path: str,
    bgcolor: str = "white",
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Persists the visualization of a prefix tree.

    :param trie: Prefix tree
    :param file_path: Destination path
    :param bgcolor: Background color (default: white)
    :param graph_title: Title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        prefix_tree = pm4py.discover_prefix_tree(dataframe, activity_key='concept:name', case_id_key='case:concept:name', timestamp_key='time:timestamp')
        pm4py.save_vis_prefix_tree(prefix_tree, 'trie.png')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.trie import visualizer as trie_visualizer

    params = _setup_parameters(fmt, bgcolor, graph_title=graph_title)
    gviz = trie_visualizer.apply(trie, parameters=params)
    return trie_visualizer.save(gviz, file_path)


def view_alignments(
    log: Union[EventLog, pd.DataFrame],
    aligned_traces: List[Dict[str, Any]],
    format: str = "png",
    graph_title: Optional[str] = None,
):
    """
    Views the alignment table as a figure.

    :param log: Event log
    :param aligned_traces: Results of an alignment
    :param format: Format of the visualization (default: png)
    :param graph_title: Title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes('tests/input_data/running-example.xes')
        net, im, fm = pm4py.discover_petri_net_inductive(log)
        aligned_traces = pm4py.conformance_diagnostics_alignments(log, net, im, fm)
        pm4py.view_alignments(log, aligned_traces, format='svg')
    """
    fmt = _extract_format(format)
    from pm4py.visualization.align_table import visualizer

    properties = {
        "format": fmt,
        "enable_graph_title": constants.DEFAULT_ENABLE_GRAPH_TITLES,
    }
    if graph_title:
        properties["enable_graph_title"] = True
        properties["graph_title"] = graph_title
    gviz = visualizer.apply(log, aligned_traces, parameters=properties)
    visualizer.view(gviz)


def save_vis_alignments(
    log: Union[EventLog, pd.DataFrame],
    aligned_traces: List[Dict[str, Any]],
    file_path: str,
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves an alignment table's figure on disk.

    :param log: Event log
    :param aligned_traces: Results of an alignment
    :param file_path: Target path
    :param graph_title: Title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes('tests/input_data/running-example.xes')
        net, im, fm = pm4py.discover_petri_net_inductive(log)
        aligned_traces = pm4py.conformance_diagnostics_alignments(log, net, im, fm)
        pm4py.save_vis_alignments(log, aligned_traces, 'output.svg')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.align_table import visualizer

    properties = {
        "format": fmt,
        "enable_graph_title": constants.DEFAULT_ENABLE_GRAPH_TITLES,
    }
    if graph_title:
        properties["enable_graph_title"] = True
        properties["graph_title"] = graph_title

    gviz = visualizer.apply(log, aligned_traces, parameters=properties)
    return visualizer.save(gviz, file_path)


def view_footprints(
    footprints: Union[Tuple[Dict[str, Any], Dict[str, Any]], Dict[str, Any]],
    format: str = "png",
    graph_title: Optional[str] = None,
):
    """
    Views the footprints as a figure.

    :param footprints: Footprints
    :param format: Format of the visualization (default: png)
    :param graph_title: Title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes('tests/input_data/running-example.xes')
        fp_log = pm4py.discover_footprints(log)
        pm4py.view_footprints(fp_log, format='svg')
    """
    fmt = _extract_format(format)
    from pm4py.visualization.footprints import visualizer as fps_visualizer

    properties = {
        "format": fmt,
        "enable_graph_title": constants.DEFAULT_ENABLE_GRAPH_TITLES,
    }
    if graph_title:
        properties["enable_graph_title"] = True
        properties["graph_title"] = graph_title

    if isinstance(footprints, dict):
        gviz = fps_visualizer.apply(footprints, parameters=properties)
    else:
        gviz = fps_visualizer.apply(
            footprints[0],
            footprints[1],
            variant=fps_visualizer.Variants.COMPARISON_SYMMETRIC,
            parameters=properties,
        )

    fps_visualizer.view(gviz)


def save_vis_footprints(
    footprints: Union[Tuple[Dict[str, Any], Dict[str, Any]], Dict[str, Any]],
    file_path: str,
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves the footprints' visualization on disk.

    :param footprints: Footprints
    :param file_path: Target path
    :param graph_title: Title of the visualization (if provided)

     .. code-block:: python3

        import pm4py

        log = pm4py.read_xes('tests/input_data/running-example.xes')
        fp_log = pm4py.discover_footprints(log)
        pm4py.save_vis_footprints(fp_log, 'output.svg')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.footprints import visualizer as fps_visualizer

    properties = {
        "format": fmt,
        "enable_graph_title": constants.DEFAULT_ENABLE_GRAPH_TITLES,
    }
    if graph_title:
        properties["enable_graph_title"] = True
        properties["graph_title"] = graph_title

    if isinstance(footprints, dict):
        gviz = fps_visualizer.apply(footprints, parameters=properties)
    else:
        gviz = fps_visualizer.apply(
            footprints[0],
            footprints[1],
            variant=fps_visualizer.Variants.COMPARISON_SYMMETRIC,
            parameters=properties,
        )

    return fps_visualizer.save(gviz, file_path)


def view_powl(
    powl: POWL,
    format: str = constants.DEFAULT_FORMAT_GVIZ_VIEW,
    bgcolor: str = "white",
    variant_str: str = "basic",
    rankdir: str = "TB",
    graph_title: Optional[str] = None,
):
    """
    Performs a visualization of a POWL model.

    Reference:
    Kourani, Humam, and Sebastiaan J. van Zelst.
    "POWL: partially ordered workflow language."
    International Conference on Business Process Management. Cham: Springer Nature Switzerland, 2023.

    :param powl: POWL model
    :param format: Format of the visualization (default: png)
    :param bgcolor: Background color (default: white)
    :param rankdir: Graph direction ("LR" or "TB")
    :param variant_str: Variant of the visualization to be used ("basic" or "net")
    :param graph_title: Title of the visualization (if provided)

     .. code-block:: python3

        import pm4py

        log = pm4py.read_xes('tests/input_data/running-example.xes')
        powl_model = pm4py.discover_powl(log)
        pm4py.view_powl(powl_model, format='svg', variant_str='basic')
        pm4py.view_powl(powl_model, format='svg', variant_str='net')
    """
    from pm4py.visualization.powl.visualizer import POWLVisualizationVariants

    variant = POWLVisualizationVariants.BASIC

    if variant_str == "basic":
        variant = POWLVisualizationVariants.BASIC
    elif variant_str == "net":
        variant = POWLVisualizationVariants.NET

    fmt = _extract_format(format)
    from pm4py.visualization.powl import visualizer as powl_visualizer

    params = _setup_parameters(fmt, bgcolor, rankdir, graph_title=graph_title)
    gviz = powl_visualizer.apply(powl, variant=variant, parameters=params)

    powl_visualizer.view(gviz, parameters=params)


def save_vis_powl(
    powl: POWL,
    file_path: str,
    bgcolor: str = "white",
    rankdir: str = "TB",
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves the visualization of a POWL model.

    Reference:
    Kourani, Humam, and Sebastiaan J. van Zelst.
    "POWL: partially ordered workflow language."
    International Conference on Business Process Management. Cham: Springer Nature Switzerland, 2023.

    :param powl: POWL model
    :param file_path: Target path
    :param bgcolor: Background color (default: white)
    :param rankdir: Graph direction ("LR" or "TB")
    :param graph_title: Title of the visualization (if provided)

     .. code-block:: python3

        import pm4py

        log = pm4py.read_xes('tests/input_data/running-example.xes')
        powl_model = pm4py.discover_powl(log)
        pm4py.save_vis_powl(powl_model, 'powl.png')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.powl import visualizer as powl_visualizer

    params = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    gviz = powl_visualizer.apply(powl, parameters=params)

    return powl_visualizer.save(gviz, file_path, parameters=params)


def view_object_graph(
    ocel: OCEL,
    graph: Set[Tuple[str, str]],
    format: str = constants.DEFAULT_FORMAT_GVIZ_VIEW,
    bgcolor: str = "white",
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    graph_title: Optional[str] = None,
):
    """
    Visualizes an object graph on the screen.

    :param ocel: Object-centric event log
    :param graph: Object graph
    :param format: Format of the visualization (if 'html' is provided, GraphvizJS is used)
    :param bgcolor: Background color (default: white)
    :param rankdir: Graph direction ("LR" or "TB")
    :param graph_title: Title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        obj_graph = pm4py.ocel_discover_objects_graph(ocel, graph_type='object_interaction')
        pm4py.view_object_graph(ocel, obj_graph, format='svg')
    """
    fmt = _extract_format(format)
    from pm4py.visualization.ocel.object_graph import (
        visualizer as obj_graph_vis,
    )

    properties = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    gviz = obj_graph_vis.apply(ocel, graph, parameters=properties)
    obj_graph_vis.view(gviz)


def save_vis_object_graph(
    ocel: OCEL,
    graph: Set[Tuple[str, str]],
    file_path: str,
    bgcolor: str = "white",
    rankdir: str = constants.DEFAULT_RANKDIR_GVIZ,
    graph_title: Optional[str] = None,
    **kwargs
):
    """
    Saves the visualization of an object graph.

    :param ocel: Object-centric event log
    :param graph: Object graph
    :param file_path: Destination path
    :param bgcolor: Background color (default: white)
    :param rankdir: Graph direction ("LR" or "TB")
    :param graph_title: Title of the visualization (if provided)

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        obj_graph = pm4py.ocel_discover_objects_graph(ocel, graph_type='object_interaction')
        pm4py.save_vis_object_graph(ocel, obj_graph, 'trial.pdf')
    """
    fmt = _extract_format(file_path)
    from pm4py.visualization.ocel.object_graph import (
        visualizer as obj_graph_vis,
    )

    properties = _setup_parameters(fmt, bgcolor, rankdir, graph_title)
    gviz = obj_graph_vis.apply(ocel, graph, parameters=properties)
    return obj_graph_vis.save(gviz, file_path)
